# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_econ.steamclient.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass

import betterproto

from .msg import UnifiedMessage


@dataclass(eq=False, repr=False)
class GetInventoryItemsWithDescriptionsRequestFilterOptions(betterproto.Message):
    assetids: list[int] = betterproto.uint64_field(1)
    currencyids: list[int] = betterproto.uint32_field(2)
    tradable_only: bool = betterproto.bool_field(3)
    marketable_only: bool = betterproto.bool_field(4)


class GetInventoryItemsWithDescriptionsRequest(UnifiedMessage, um_name="Econ.GetInventoryItemsWithDescriptions"):
    steamid: int = betterproto.fixed64_field(1)
    appid: int = betterproto.uint32_field(2)
    contextid: int = betterproto.uint64_field(3)
    get_descriptions: bool = betterproto.bool_field(4)
    for_trade_offer_verification: bool = betterproto.bool_field(10)
    language: str = betterproto.string_field(5)
    filters: GetInventoryItemsWithDescriptionsRequestFilterOptions = betterproto.message_field(6)
    start_assetid: int = betterproto.uint64_field(8)
    count: int = betterproto.int32_field(9)


@dataclass(eq=False, repr=False)
class Asset(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    contextid: int = betterproto.uint64_field(2)
    assetid: int = betterproto.uint64_field(3)
    classid: int = betterproto.uint64_field(4)
    instanceid: int = betterproto.uint64_field(5)
    currencyid: int = betterproto.uint32_field(6)
    amount: int = betterproto.int64_field(7)
    missing: int = betterproto.bool_field(8)
    est_usd: int = betterproto.int64_field(9)


@dataclass(eq=False, repr=False)
class ItemDescriptionLine(betterproto.Message):
    type: str = betterproto.string_field(1)
    value: str = betterproto.string_field(2)
    color: str = betterproto.string_field(3)
    label: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class ItemAction(betterproto.Message):
    link: str = betterproto.string_field(1)
    name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ItemTag(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    category: str = betterproto.string_field(2)
    internal_name: str = betterproto.string_field(3)
    localized_category_name: str = betterproto.string_field(4)
    localized_tag_name: str = betterproto.string_field(5)
    color: str = betterproto.string_field(6)


@dataclass(eq=False, repr=False)
class ItemDescription(betterproto.Message):
    appid: int = betterproto.int32_field(1)
    classid: int = betterproto.uint64_field(2)
    instanceid: int = betterproto.uint64_field(3)
    currency: bool = betterproto.bool_field(4)
    background_color: int = betterproto.string_field(5)
    icon_url: str = betterproto.string_field(6)
    icon_url_large: str = betterproto.string_field(7)
    descriptions: list[ItemDescriptionLine] = betterproto.message_field(8)
    tradable: bool = betterproto.bool_field(9)
    actions: list[ItemAction] = betterproto.message_field(10)
    owner_descriptions: list[ItemDescriptionLine] = betterproto.message_field(11)
    owner_actions: list[ItemAction] = betterproto.message_field(12)
    fraudwarnings: list[str] = betterproto.string_field(13)
    name: str = betterproto.string_field(14)
    name_color: str = betterproto.string_field(15)
    type: str = betterproto.string_field(16)
    market_name: str = betterproto.string_field(17)
    market_hash_name: str = betterproto.string_field(18)
    market_fee: str = betterproto.string_field(19)
    market_fee_app: int = betterproto.int32_field(28)
    contained_item: "ItemDescription" = betterproto.message_field(20)
    market_actions: list[ItemAction] = betterproto.message_field(21)
    commodity: bool = betterproto.bool_field(22)
    market_tradable_restriction: int = betterproto.int32_field(23)
    market_marketable_restriction: int = betterproto.int32_field(24)
    marketable: bool = betterproto.bool_field(25)
    tags: list[ItemTag] = betterproto.message_field(26)
    item_expiration: str = betterproto.string_field(27)
    market_buy_country_restriction: str = betterproto.string_field(30)
    market_sell_country_restriction: str = betterproto.string_field(31)


class GetInventoryItemsWithDescriptionsResponse(UnifiedMessage, um_name="Econ.GetInventoryItemsWithDescriptions"):
    assets: list[Asset] = betterproto.message_field(1)
    descriptions: list[ItemDescription] = betterproto.message_field(2)
    missing_assets: list[Asset] = betterproto.message_field(3)
    more_items: bool = betterproto.bool_field(4)
    last_assetid: int = betterproto.uint64_field(5)
    total_inventory_count: int = betterproto.uint32_field(6)


class GetTradeOfferAccessTokenRequest(UnifiedMessage, um_name="Econ.GetTradeOfferAccessToken"):
    generate_new_token: bool = betterproto.bool_field(1)


class GetTradeOfferAccessTokenResponse(UnifiedMessage, um_name="Econ.GetTradeOfferAccessToken"):
    trade_offer_access_token: str = betterproto.string_field(1)


class GetItemShopOverlayAuthUrlRequest(UnifiedMessage, um_name="Econ.ClientGetItemShopOverlayAuthURL"):
    return_url: str = betterproto.string_field(1)


class GetItemShopOverlayAuthUrlResponse(UnifiedMessage, um_name="Econ.ClientGetItemShopOverlayAuthURL"):
    url: str = betterproto.string_field(1)


class GetAssetClassInfoRequest(UnifiedMessage, um_name="Econ.GetAssetClassInfo"):
    language: str = betterproto.string_field(1)
    appid: int = betterproto.uint32_field(2)
    classes: "list[GetAssetClassInfoRequestClass]" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class GetAssetClassInfoRequestClass(betterproto.Message):
    classid: int = betterproto.uint64_field(1)
    instanceid: int = betterproto.uint64_field(2)


class GetAssetClassInfoResponse(UnifiedMessage, um_name="Econ.GetAssetClassInfo"):
    descriptions: "list[ItemDescription]" = betterproto.message_field(1)


class GetTradeOfferRequest(UnifiedMessage, um_name="Econ.GetTradeOffer"):
    tradeofferid: int = betterproto.uint64_field(1)


class GetTradeOfferResponse(UnifiedMessage, um_name="Econ.GetTradeOffer"):
    tradeofferid: int = betterproto.uint64_field(1)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import errno
import logging
import lzma
import os.path
import struct
import sys
from base64 import b64decode
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from io import BytesIO
from operator import attrgetter, methodcaller
from typing import TYPE_CHECKING, Any, Final, Literal, TypeGuard, cast, overload
from zipfile import BadZipFile, ZipFile
from zlib import crc32

from aiohttp.streams import AsyncStreamReaderMixin
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from multidict import MultiDict

from . import utils
from ._const import MISSING, URL, VDF_LOADS, VDFDict, impl_eq_via_id
from .app import PartialApp
from .enums import (
    AppType,
    BillingType,
    ContentDescriptor,
    DepotFileFlag,
    Language,
    LicenseType,
    PackageStatus,
    ReviewType,
)
from .id import ID
from .models import CDNAsset, _IOMixin
from .package import PartialPackage
from .protobufs.content_manifest import Metadata, Payload, PayloadFileMapping, PayloadFileMappingChunkData, Signature
from .tag import Category, Genre, Tag
from .types.id import AppID, DepotID, ManifestID
from .utils import DateTime, cached_slot_property

if sys.platform == "win32":
    from pathlib import PureWindowsPath as PurePathBase
else:
    from pathlib import PurePosixPath as PurePathBase


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Generator, Mapping, Sequence
    from datetime import datetime

    from _typeshed import StrPath
    from typing_extensions import Never, Self
    from yarl import URL as URL_

    from .protobufs import app_info
    from .state import ConnectionState
    from .types import manifest, manifest as manifest_
    from .types.vdf import VDFInt


__all__ = (
    "Manifest",
    "ManifestPath",
    "Branch",
    "ManifestInfo",
    "PrivateManifestInfo",
    "HeadlessDepot",
    "Depot",
    "AppInfo",
    "PackageInfo",
)

log = logging.getLogger(__name__)


def unzip(data: bytes, /) -> bytes:
    if data[:2] == b"VZ":
        if data[-2:] != b"zv":
            raise RuntimeError(f"VZ: Invalid footer: {data[-2:]!r}")
        if data[2] != 97:
            raise RuntimeError(f"VZ: Invalid version: {data[2]!r}")

        filters = (lzma._decode_filter_properties(lzma.FILTER_LZMA1, data[7:12]),)  # type: ignore
        decompressor = lzma.LZMADecompressor(lzma.FORMAT_RAW, filters=filters)
        checksum, decompressed_size = struct.unpack("<II", data[-10:-2])
        data = decompressor.decompress(data[12:-9], max_length=decompressed_size)

        if crc32(data) != checksum:
            raise RuntimeError("VZ: CRC32 checksum doesn't match for decompressed data")

    else:
        try:
            with ZipFile(BytesIO(data)) as zf:
                data = zf.read(zf.filelist[0])
        except BadZipFile:
            pass

    return data


@dataclass(slots=True)
class ManifestPathIO(AsyncStreamReaderMixin):
    _path: ManifestPath
    _key: bytes
    _chunk_idx: int = 0
    _buffer: bytearray = field(default_factory=bytearray)

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, *args: object) -> None:
        self._buffer.clear()
        self._chunk_idx = len(self._path.chunks)

    async def _get(self, chunk: PayloadFileMappingChunkData) -> None:
        data = await self._path._manifest.server.get(f"depot/{self._path._manifest.depot_id}/chunk/{chunk.sha.hex()}")
        self._chunk_idx += 1
        self._buffer += unzip(utils.symmetric_decrypt(data, self._key))

    async def read(self, n: int = -1, /) -> bytes:
        if n == -1:
            async with self:
                self._buffer += b"".join(
                    unzip(utils.symmetric_decrypt(data, self._key))
                    for data in await asyncio.gather(
                        *(
                            self._path._manifest.server.get(
                                f"depot/{self._path._manifest.depot_id}/chunk/{chunk.sha.hex()}"
                            )
                            for chunk in self._path.chunks[self._chunk_idx :]
                        )
                    )
                )
                return bytes(self._buffer)

        for chunk in self._path.chunks[self._chunk_idx :]:
            if len(self._buffer) >= n:
                break
            await self._get(chunk)

        found = bytes(self._buffer[:n])
        self._buffer = self._buffer[n:]
        return found

    async def read1(self, n: int = -1, /) -> bytes:
        if len(self._buffer) < n or n == -1:
            await self._get(self._path.chunks[self._chunk_idx])

        found = bytes(self._buffer[:n])
        self._buffer = self._buffer[n:]
        return found

    async def readuntil(self, separator: bytes = b"\n", /) -> bytes:
        while (idx := self._buffer.find(separator)) == -1 and self._chunk_idx != len(self._path.chunks):
            await self._get(self._path.chunks[self._chunk_idx])

        found = bytes(self._buffer[: idx + len(separator)])
        self._buffer = self._buffer[idx + len(separator) :]
        return found

    async def readline(self) -> bytes:
        return await self.readuntil()

    async def readany(self) -> bytes:
        if self._buffer:
            data = bytes(self._buffer)
            self._buffer.clear()
            return data

        return await self.read1()

    async def readchunk(self) -> tuple[bytes, Literal[True]]:
        return await self.read1(), True

    async def readexactly(self, n: int, /) -> bytes:
        data = await self.read(n)
        if len(data) != n:
            raise asyncio.IncompleteReadError(data, n)

        return data

    def read_nowait(self, n: int = -1, /) -> bytes:
        content = bytes(self._buffer[:n]) if n != -1 else bytes(self._buffer)
        self._buffer = self._buffer[n:]
        return content


def _manifest_parts(filename: str, /) -> list[str]:
    return filename.rstrip("\x00 \n\t").split("\\")


class ManifestPath(PurePathBase):
    """A :class:`pathlib.PurePath` subclass representing a binary file in a Manifest. This class is broadly compatible
    with :class:`pathlib.Path`.

    .. container:: operations

        .. describe:: x == y

            Checks if two paths are equal.

        .. describe:: hash(x)

            Hashes this path.

        .. describe:: x < y

            Checks if one path's parts is less than the other.
    """

    __slots__ = ("_manifest", "_mapping", "_flags_cs")

    _manifest: Manifest
    _mapping: PayloadFileMapping

    if sys.version_info < (3, 12):

        def __new__(cls, *args: StrPath, manifest: Manifest, mapping: PayloadFileMapping | None = None) -> Self:
            # super().__new__ breaks
            self: Self = super()._from_parts(_manifest_parts(mapping.filename) if mapping is not None else args)  # type: ignore
            self._manifest = manifest
            if mapping is not None:
                self._mapping = mapping
            return self

        def with_segments(self, *args: StrPath) -> ManifestPath:
            return self._select_from_manifest(self._from_parts(self.parts + tuple(os.fspath(arg) for arg in args)))

        def _select_from_manifest(self, new_self: ManifestPath) -> ManifestPath:
            try:
                # try and return the actual path if exists
                return self._manifest._paths[new_self.parts]
            except KeyError:
                # else attach the manifest and return, this will not support most operations
                new_self._manifest = self._manifest
                return new_self

        def _from_parts(self, args: tuple[str, ...]) -> ManifestPath:
            new_self = super()._from_parts(args)  # type: ignore
            return self._select_from_manifest(new_self)

        def _from_parsed_parts(self, drv: str, root: str, parts: tuple[str, ...]) -> ManifestPath:
            new_self = super()._from_parsed_parts(drv, root, parts)  # type: ignore
            return self._select_from_manifest(new_self)

        @property
        def parents(self) -> tuple[Self, ...]:
            """A tuple of this path's logical parents."""
            path = self
            parent = self.parent
            parents: list[Self] = []
            while path != parent:
                parents.append(parent)
                path, parent = parent, parent.parent
            return tuple(parents)

    else:

        def __init__(
            self,
            *args: StrPath,
            manifest: Manifest,
            mapping: PayloadFileMapping | None = None,
        ):
            super().__init__(*_manifest_parts(mapping.filename) if mapping is not None else args)
            self._manifest = manifest
            if mapping is not None:
                self._mapping = mapping

        def with_segments(self, *args: StrPath) -> Self:
            new_self = self.__class__(*args, manifest=self._manifest)
            try:
                # try and return the actual path if exists
                return self._manifest._paths[new_self.parts]
            except KeyError:
                return new_self

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {str(self)!r}>"

    if not TYPE_CHECKING:

        def __getattr__(self, name: str) -> Never:
            if name in self.__annotations__:  # give a more helpful error
                raise ValueError("Attempting operations on a non-existent file")
            raise AttributeError(f"{self.__class__.__name__!r} object has no attribute {name!r}", name=name, obj=self)

    @property
    def size(self) -> int:
        """This path's size in bytes."""
        return self._mapping.size

    @property
    def chunks(self) -> Sequence[PayloadFileMappingChunkData]:
        """A read-only version of this path's chunks instances."""
        return self._mapping.chunks

    @utils.cached_slot_property
    def flags(self) -> DepotFileFlag:
        """This path's flags."""
        return DepotFileFlag.try_value(self._mapping.flags)

    @property
    def sha_content(self) -> bytes:
        """This path's SHA1 content hash."""
        return self._mapping.sha_content

    @property
    def sha_filename(self) -> bytes:
        """This path's SHA1 filename hash."""
        return self._mapping.sha_filename

    def is_dir(self) -> bool:
        """Whether the path is a directory."""
        return self.flags & DepotFileFlag.Directory > 0

    def is_symlink(self) -> bool:
        """Whether this path is a symlink."""
        return self.flags & DepotFileFlag.Symlink > 0

    def is_file(self) -> bool:
        """Whether this path is a file or symlink."""
        return not self.is_dir()

    def is_executable(self) -> bool:
        """Whether this file is executable."""
        return self.flags & DepotFileFlag.Executable > 0

    def is_hidden(self) -> bool:
        """Whether this file is hidden."""
        return self.flags & DepotFileFlag.Hidden > 0

    def readlink(self) -> Self:
        """If this path is a symlink where it points to. Similar to :meth:`pathlib.Path.readlink`

        Raises
        ------
        OSError
            This path isn't a symlink.
        """
        if not self.is_symlink():
            raise OSError(errno.EINVAL, os.strerror(errno.EINVAL), str(self))

        link_parts = tuple(_manifest_parts(self._mapping.linktarget))
        return self._manifest._paths[link_parts]

    @overload
    def resolve(self, *, strict: bool = False) -> Self:  # type: ignore
        ...

    def resolve(self, *, strict: bool = False, _follow_symlinks: bool = True) -> Self:
        """Return the canonical path of the symbolic link, eliminating any symbolic links encountered in the path.
        Similar to :meth:`pathlib.Path.resolve`

        Parameters
        ----------
        strict
            Whether to raise an error if a path doesn't exist.

        Raises
        ------
        FileNotFoundError
            If ``strict`` is ``True`` and the path doesn't exist.
        RuntimeError
            If a recursive path is detected.
        """
        new_parts: list[str] = []
        seen = set[tuple[str, ...]]()
        idx = 0
        raw_parts = list(self.parts)
        if not raw_parts:
            raise RuntimeError("Cannot resolve empty path")

        for part in raw_parts:
            match part:
                case "." | "":
                    idx += 1
                    continue
                case "..":
                    raw_parts.insert(idx + 1, raw_parts[idx - 1])
                    idx += 1
                    continue
            new_parts.append(part)

            path = self.with_segments(*new_parts)

            if not hasattr(path, "_mapping"):
                if strict:
                    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(self))
            elif path.is_symlink() and _follow_symlinks:
                new_parts = _manifest_parts(path._mapping.linktarget)

            if (tuple_parts := tuple(new_parts)) in seen:
                raise RuntimeError("Recursive path detected. Cannot resolve")
            seen.add(tuple_parts)
            idx += 1

        return path  # type: ignore  # cannot be unbound

    def exists(self, *, follow_symlinks: bool = True) -> bool:
        """Return whether this path exists. Similar to :meth:`pathlib.Path.exists`."""
        return hasattr(
            self.resolve(
                strict=False,
                _follow_symlinks=follow_symlinks,  # type: ignore
            ),
            "_mapping",
        )

    def iterdir(self) -> Generator[Self, None, None]:
        """Iterate over this path. Similar to :meth:`pathlib.Path.iterdir`."""
        for path in self._manifest._paths.values():
            if path.parent == self:
                yield path

    def walk(
        self, *, top_down: bool = True, follow_symlinks: bool = False
    ) -> Generator[tuple[Self, list[str], list[str]], None, None]:
        """Walk this path. Similar to :meth:`pathlib.Path.walk`.

        Parameters
        ----------
        top_down
            Whether to walk top down or bottom up.
        follow_symlinks
            Whether to follow symlinks.

        Note
        ----
        Unlike :meth:`pathlib.Path.walk`, this method does not have a ``on_error`` parameter as it should never error.

        Yields
        ------
        The path currently being traversed, directories and files (``(dirpath, dirnames, filenames)``).
        """

        paths: list[Self | tuple[Self, list[str], list[str]]] = [self]

        while paths:
            path = paths.pop()
            if isinstance(path, tuple):
                yield path
                continue

            dirnames: list[str] = []
            filenames: list[str] = []
            for entry in path.iterdir():
                is_dir = entry.is_dir() if follow_symlinks else entry.is_dir() and not entry.is_symlink()
                (dirnames if is_dir else filenames).append(entry.name)

            if top_down:
                yield path, dirnames, filenames
            else:
                paths.append((path, dirnames, filenames))

            paths += [path._make_child_relpath(d) for d in reversed(dirnames)]  # type: ignore

    def glob(self, pattern: str, /) -> Generator[Self, None, None]:
        """Perform a glob operation on this path. Similar to :meth:`pathlib.Path.glob`."""
        if not pattern:
            raise ValueError(f"Unacceptable pattern: {pattern!r}")

        yield from filter(
            methodcaller("match", f"{self.as_posix().removesuffix('/')}/{pattern}"), self._manifest._paths.values()
        )

    def rglob(self, pattern: str, /) -> Generator[Self, None, None]:
        """Perform a recursive glob operation on this path. Similar to :meth:`pathlib.Path.rglob`."""
        yield from self.glob(f"**/{pattern}")

    @asynccontextmanager
    async def open(self) -> AsyncGenerator[ManifestPathIO, None]:
        """Reads the contents of the file.

        Raises
        ------
        IsADirectoryError
            This path is a directory and cannot be opened.
        RuntimeError
            The depot cannot be decrypted as no key for its manifest was found.
        """
        if self.is_dir():
            raise IsADirectoryError(errno.EISDIR, os.strerror(errno.EISDIR), str(self))

        key = self._manifest._key
        if key is None:
            raise RuntimeError("Cannot decrypt this depot as we have no key.")

        async with ManifestPathIO(self, key) as file:
            yield file

    read_bytes = _IOMixin.read
    """Read the contents of the file. Similar to :meth:`pathlib.Path.read_bytes`"""
    media = _IOMixin.media
    save = _IOMixin.save

    async def read_text(self, encoding: str = MISSING, errors: str = MISSING) -> str:
        """Read the contents of the file. Similar to :meth:`pathlib.Path.read_text`"""
        contents = await self.read_bytes()
        if encoding is MISSING:
            return contents.decode() if errors is MISSING else contents.decode(errors=errors)

        return contents.decode(encoding) if errors is MISSING else contents.decode(encoding, errors)

    async def read_vdf(self, encoding: str = MISSING, errors: str = MISSING) -> VDFDict:
        """Read the contents of the file into a VDFDict."""
        return VDF_LOADS(await self.read_text(encoding, errors))


PAYLOAD_MAGIC: Final = 0x71F617D0
METADATA_MAGIC: Final = 0x1F4812BE
SIGNATURE_MAGIC: Final = 0x1B81B817
END_OF_MANIFEST_MAGIC: Final = 0x32C415AB


@impl_eq_via_id
class Manifest:
    """Represents a manifest which is a collection of files included with a depot build on Steam's CDN.

    .. container:: operations

        .. describe:: x == y

            Checks if two manifests are equal.

        .. describe:: len(x)

            Returns the number of files this manifest holds.

        .. describe:: bytes(x)

            Returns the manifest's binary data.

            Note
            ----
            This will be the uncompressed data.
    """

    __slots__ = (
        "name",
        "app",
        "server",
        "_key",
        "_metadata",
        "_payload",
        "_signature",
        "_state",
        "_cs_paths",
        "_cs_created_at",
    )

    def __init__(self, state: ConnectionState, server: ContentServer, app_id: AppID, data: bytes):
        self._state = state
        self.name: str | None = None
        """The name of the manifest."""
        self.app = PartialApp(state, id=app_id)
        """The app that this manifest was fetched from."""
        self.server = server
        """The content server that this manifest was fetched from."""
        self._key: bytes | None = None

        with utils.StructIO(unzip(data)) as io:
            if io.read_u32() != PAYLOAD_MAGIC:
                raise RuntimeError("Expecting protobuf payload")
            length = io.read_u32()
            self._payload = Payload().parse(io.read(length))

            if io.read_u32() != METADATA_MAGIC:
                raise RuntimeError("Expecting protobuf metadata")
            length = io.read_u32()
            self._metadata = Metadata().parse(io.read(length))

            if io.read_u32() != SIGNATURE_MAGIC:
                raise RuntimeError("Expecting protobuf signature")
            length = io.read_u32()
            self._signature = Signature().parse(io.read(length))

            if io.read_u32() != END_OF_MANIFEST_MAGIC:
                raise RuntimeError("Expecting end of manifest")

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id} depot_id={self.depot_id}>"

    def __bytes__(self) -> bytes:
        with utils.StructIO() as io:
            io.write_u32(PAYLOAD_MAGIC)
            payload = bytes(self._payload)
            io.write_u32(len(payload))
            io.write(payload)

            io.write_u32(METADATA_MAGIC)
            metadata = bytes(self._metadata)
            io.write_u32(len(metadata))
            io.write(metadata)

            io.write_u32(SIGNATURE_MAGIC)
            signature = bytes(self._signature)
            io.write_u32(len(signature))
            io.write(signature)

            io.write_u32(END_OF_MANIFEST_MAGIC)
            return io.buffer

    def __len__(self) -> int:
        return len(self._payload.mappings)

    @cached_slot_property("_cs_paths")
    def _paths(self) -> dict[tuple[str, ...], ManifestPath]:
        return {("/",): ManifestPath(manifest=self, mapping=PayloadFileMapping("/", flags=DepotFileFlag.Directory))} | {
            (path := ManifestPath(manifest=self, mapping=mapping)).parts: path for mapping in self._payload.mappings
        }

    @property
    def paths(self) -> Sequence[ManifestPath]:
        """The depot's files."""
        return list(self._paths.values())

    @property
    def root(self) -> ManifestPath:
        """The depot's root directory."""
        return self._paths[("/",)]

    @property
    def id(self) -> ManifestID:
        """The ID of this manifest. This is randomly generated by Steam."""
        return ManifestID(self._metadata.gid_manifest)

    @property
    def depot_id(self) -> DepotID:
        """The ID of this manifest's depot."""
        return DepotID(self._metadata.depot_id)

    @cached_slot_property("_cs_created_at")
    def created_at(self) -> datetime:
        """The time at which the depot was created at."""
        return DateTime.from_timestamp(self._metadata.creation_time)

    @property
    def size_original(self) -> int:
        """The size of the original depot file."""
        return self._metadata.cb_disk_original

    @property
    def size_compressed(self) -> int:
        """The size of the compressed depot file."""
        return self._metadata.cb_disk_compressed

    @property
    def compressed(self) -> bool:
        """Whether the depot is compressed."""
        return self.size_original != self.size_compressed if self.size_compressed else False


@dataclass(slots=True)
class ContentServer(ID):  # is there any point having this inherit steamid?
    _state: ConnectionState
    url: URL_
    weighted_load: float

    async def get(self, path: str) -> bytes:
        async with self._state.http._session.get(self.url / path) as r:
            return await r.read()

    async def fetch_manifest(
        self,
        app_id: AppID,
        id: ManifestID,
        depot_id: DepotID,
        name: str | None = None,
        branch: str = "public",
        password_hash: str = "",
    ) -> Manifest:
        branch = branch if branch != "public" else ""
        code = await self._state.fetch_manifest_request_code(id, depot_id, app_id, branch, password_hash)
        data = await self.get(f"depot/{depot_id}/manifest/{id}/5{f'/{code}' if code else ''}")

        manifest = Manifest(self._state, self, app_id, data)
        encrypted = manifest._metadata.filenames_encrypted
        if encrypted:
            key = manifest._key = await self._state.fetch_depot_key(app_id, depot_id)
        for mapping in manifest._payload.mappings:
            if encrypted:
                mapping.filename = utils.symmetric_decrypt(b64decode(mapping.filename), key).decode()  # type: ignore # key is never unbound
                if mapping.linktarget:
                    mapping.linktarget = utils.symmetric_decrypt(b64decode(mapping.linktarget), key).decode()  # type: ignore

            mapping.chunks.sort(key=attrgetter("offset"), reverse=False)

        manifest.name = name
        return manifest


@dataclass(slots=True)
class Branch:
    """Represents a branch on for a Steam app. Branches are specific builds of an application that have made available
    publicly or privately through Steam.

    Read more on `steamworks <https://partner.steamgames.com/doc/store/application/branches>`_.
    """

    name: str
    """The name of the branch."""
    build_id: int
    """
    The branch's build ID. This is a globally incrementing number. Build IDs are updated when a new build of an
    application is pushed.
    """
    updated_at: datetime | None
    """The time this branch was last updated."""
    password_required: bool
    """Whether a password is required to access this branch."""
    description: str | None
    """This branch's description."""
    depots: list[Depot] = field(default_factory=list)
    """This branch's depots."""
    password: str | None = None
    """This branch's password."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} build_id={self.build_id}>"

    @property
    def manifests(self) -> list[ManifestInfo]:
        """This branch's manifests."""
        return [depot.manifest for depot in self.depots]

    async def fetch_manifests(self) -> list[Manifest]:
        """Fetch this branch's manifests. Similar to :meth:`PartialApp.manifests`."""
        return await asyncio.gather(*(manifest.fetch() for manifest in self.manifests))


@dataclass(slots=True)
class ManifestInfo:
    """Represents information about a manifest."""

    _state: ConnectionState
    id: ManifestID
    """The manifest's ID."""
    branch: Branch
    """The branch this manifest is for."""
    size: int | None
    download_size: int | None
    depot: Depot = field(init=False)
    """The depot this manifest is for."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id}>"

    @property
    def name(self) -> str | None:
        """The manifest's name."""
        return self.depot.name

    async def fetch(self) -> Manifest:
        """Resolves this manifest info into a full :class:`Manifest`."""
        return await self._state.fetch_manifest(self.depot.app.id, self.id, self.depot.id, self.name)


class PrivateManifestInfo(ManifestInfo):
    __slots__ = ("encrypted_id", "_id_cs")

    def __init__(self, state: ConnectionState, encrypted_id: str, branch: Branch):
        self._state = state
        self.encrypted_id = encrypted_id
        self.branch = branch

    @cached_slot_property
    def id(self) -> ManifestID:  # type: ignore
        if self.branch.password is None:
            raise ValueError("Cannot access the id of this depot as the password is not set.")
        cipher = Cipher(algorithms.AES(self.branch.password.encode("UTF-8")), modes.ECB())
        decryptor = cipher.decryptor()
        to_unpack = utils.unpad(decryptor.update(bytes.fromhex(self.encrypted_id) + decryptor.finalize()))
        return ManifestID(int.from_bytes(to_unpack, byteorder="little"))

    @staticmethod
    def _get_id(depots: manifest.Depot, branch: Branch) -> VDFInt | None:
        try:
            return depots["encryptedmanifests"][branch.name]["encrypted_gid_2"]
        except KeyError:
            return None


@dataclass(slots=True)
@impl_eq_via_id
class HeadlessDepot:
    """Represents a depot without a branch."""

    id: DepotID
    """The depot's ID."""
    name: str | None
    """The depot's name."""
    app: AppInfo
    """The depot's app."""
    max_size: int | None
    """The depot's maximum size."""
    config: MultiDict[str]
    """The depot's configuration settings."""
    shared_install: bool | None
    """Whether this depot supports shared installs"""
    system_defined: bool | None
    """Whether this depot is system defined."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id} app={self.app!r}>"


@dataclass(repr=False, slots=True)
class Depot(HeadlessDepot):
    """Represents a depot which is a grouping of files.

    Read more on `steamworks <https://partner.steamgames.com/doc/store/application/depots>`_.
    """

    manifest: ManifestInfo
    """The depot's associated manifest."""
    branch: Branch
    """The branch the depot is for."""


class ProductInfo:
    __slots__ = ()
    _state: ConnectionState
    sha: str
    """The product info's SHA."""
    size: int
    """The product info's size."""
    change_number: int
    """The product info's change number."""

    def __init__(
        self,
        state: ConnectionState,
        proto: (
            app_info.CMsgClientPicsProductInfoResponseAppInfo | app_info.CMsgClientPicsProductInfoResponsePackageInfo
        ),
        **kwargs: Any,
    ):
        super().__init__(state, **kwargs)  # type: ignore  # this is a mixin
        self.sha = proto.sha.hex()
        self.size = proto.size
        self.change_number = proto.change_number


def is_depot(item: tuple[str, Any]) -> TypeGuard[tuple[str, manifest.Depot]]:
    try:
        int(item[0])  # only the integer keys are depots
    except ValueError:
        return False
    else:
        return True


def maybe_get(map: Mapping[str, str | None] | str, key: str) -> int | None:
    if isinstance(map, str):
        return int(map)
    if (value := map.get(key)) is not None:
        return int(value)


class AppInfo(ProductInfo, PartialApp[str]):
    """Represents a collection of information on an app."""

    __slots__ = (
        "sha",
        "size",
        "change_number",
        "_branches",
        "headless_depots",
        "type",
        "has_adult_content",
        "has_adult_content_sex",
        "has_adult_content_violence",
        "market_presence",
        "workshop_visible",
        "community_hub_visible",
        "controller_support",
        "publishers",
        "developers",
        "supported_languages",
        "language_support",
        "tags",
        "categories",
        "content_descriptors",
        "genres",
        "created_at",
        "review_score",
        "review_percentage",
        "metacritic_name",
        "metacritic_score",
        "metacritic_url",
        "partial_dlc",
        "icon",
        "logo",
        "website_url",
        "parent",
        "demo_of_app",
        "_stats_visible",
        "_free",
        "_on_windows",
        "_on_mac_os",
        "_on_linux",
    )
    name: str

    def __init__(
        self,
        state: ConnectionState,
        data: manifest.AppInfo,
        proto: app_info.CMsgClientPicsProductInfoResponseAppInfo,
    ):
        common = data["common"]
        extended = data.get("extended", {})
        super().__init__(state, proto, id=proto.appid, name=common["name"])
        self.type = AppType.from_str(common["type"])
        """The app's type."""
        self.has_adult_content = common.get("has_adult_content", "0") == "1"
        """Whether this app has adult content according to Steam."""
        self.has_adult_content_violence = common.get("has_adult_content_violence", "0") == "1"
        """Whether this app has adult sexual content according to Steam."""
        self.has_adult_content_sex = common.get("has_adult_content_sex", "0") == "1"
        """Whether this app has adult violence according to Steam."""
        self.market_presence = common.get("market_presence", "0") == "1"
        """Whether this app has a market presence."""
        self.workshop_visible = common.get("workshop_visible", "0") == "1"
        """Whether this app has a workshop visible."""
        self.community_hub_visible = common.get("community_hub_visible", "0") == "1"
        """Whether this app has a content hub visible."""
        self._stats_visible = common.get("community_visible_stats", "0") == "1"
        self._free = extended.get("isfreeapp", "0") == "1"
        self.content_descriptors = [
            ContentDescriptor.try_value(int(id)) for id in common.get("content_descriptors", MultiDict()).values()
        ]
        """This app's content descriptors."""

        self.controller_support = common.get("controller_support", "none")
        """This app's level of controller support."""

        self.publishers: list[str] = [
            publisher["name"]
            for publisher in common.get("associations", MultiDict()).values()
            if publisher["type"] == "publisher"
        ]
        """This app's publishers."""
        self.developers: list[str] = [
            developer["name"]
            for developer in common.get("associations", MultiDict()).values()
            if developer["type"] == "developer"
        ]
        """This app's developers."""

        self.supported_languages = [
            Language.from_str(language)
            for language, value in common.get("languages", MultiDict()).items()
            if value == "1"
        ]
        """This app's supported languages."""

        self.language_support = (
            {Language.from_str(language): support for language, support in supported_languages.items()}
            if isinstance(supported_languages := common.get("supported_languages"), MultiDict)
            else {}
        )
        """This app's language support."""

        self.categories = [
            Category(state, id=int(name.removeprefix("category_"))) for name in common.get("category", ())
        ]
        """This app's supported categories."""
        self.tags = [Tag(state, id=int(tag_id)) for tag_id in common.get("store_tags", MultiDict()).values()]
        """This app's supported tags."""
        self.genres = [Genre(state, id=int(genre_id)) for genre_id in common.get("genres", MultiDict()).values()]
        """This app's supported genres."""

        self.created_at = (
            DateTime.from_timestamp(int(common["steam_release_date"])) if "steam_release_date" in common else None
        )
        """The time this app was created."""
        self.review_score = ReviewType.try_value(int(common.get("review_score", 0)))
        """This app's review score."""
        self.review_percentage = int(common.get("review_percentage", 0))
        """This app's review percentage."""
        self.metacritic_name = common.get("metacritic_name", None)
        """This app's metacritic name."""
        self.metacritic_score = int(common["metacritic_score"]) if "metacritic_score" in common else None
        """This app's metacritic score."""
        self.metacritic_url = common.get("metacritic_fullurl", None)
        """This app's metacritic URL."""

        dlc = extended.get("listofdlc", "")
        self.partial_dlc = [PartialApp(state, id=int(id)) for id in dlc.split(",")] if dlc else []
        """This app's downloadable content."""

        os_list = common.get("oslist", "")
        self._on_windows = "windows" in os_list
        self._on_mac_os = "macos" in os_list
        self._on_linux = "linux" in os_list

        self.icon = (
            CDNAsset(state, f"{URL.CDN}/steam/apps/{self.id}/{common['icon']}.jpg") if "icon" in common else None
        )
        """This app's icon."""
        self.logo = (
            CDNAsset(state, f"{URL.CDN}/steam/apps/{self.id}/{common['logo']}.jpg") if "logo" in common else None
        )
        """This app's logo."""
        self.website_url = extended.get("homepage")
        """This app's website URL."""
        self.parent = PartialApp(state, id=int(common["parent"])) if "parent" in common else None
        """This app's parent."""
        self.demo_of_app = PartialApp(state, id=int(extended["demoofappid"])) if "demoofappid" in extended else None

        depots: manifest.Depot = data.get("depots", MultiDict())  # type: ignore
        self._branches: dict[str, Branch] = {}

        for name, value in depots.get("branches", MultiDict()).items():
            try:
                build_id = int(value["buildid"])
            except KeyError:
                log.debug("Got a branch %s with no build id, discarding", name)
            else:
                self._branches[name] = Branch(
                    name=name,
                    build_id=build_id,
                    updated_at=DateTime.from_timestamp(int(value["timeupdated"])) if "timeupdated" in value else None,
                    password_required=bool(int(value.get("pwdrequired", False))),
                    description=value.get("description") or None,
                )
        self.headless_depots: Sequence[HeadlessDepot] = []
        """This app's headless depots."""

        for key, depot in filter(is_depot, depots.items()):
            id = DepotID(int(key))
            if not depot:
                depot = cast("manifest_.Depot", MultiDict(name=f"{self.name} ({id}) Depot"))
            name = depot.get("name")
            config = depot.get("config", MultiDict())
            max_size = int(depot["maxsize"]) if "maxsize" in depot else None
            shared_install = bool(int(shared_install)) if (shared_install := depot.get("sharedinstall")) else None
            system_defined = bool(int(system_defined)) if (system_defined := depot.get("system_defined")) else None
            try:
                manifests = depot["manifests"]
            except KeyError:
                self.headless_depots.append(
                    HeadlessDepot(
                        id=id,
                        name=name,
                        config=config,
                        app=self,
                        max_size=max_size,
                        shared_install=shared_install,
                        system_defined=system_defined,
                    )
                )
            else:
                if "public" not in manifests:
                    return  # we don't own the app, no branch info available
                public_manifest_info = manifests["public"]
                public_id = ManifestID(
                    int(public_manifest_info if isinstance(public_manifest_info, str) else public_manifest_info["gid"])
                )
                for branch_name, manifest_info in manifests.items():
                    try:
                        branch = self._branches[branch_name]
                    except KeyError:
                        log.debug("Got a manifest %s that has no associated branch", manifest_info)
                        continue
                    manifest_id = manifest_info if isinstance(manifest_info, str) else manifest_info["gid"]
                    if not branch.password_required:
                        manifest = ManifestInfo(
                            state,
                            ManifestID(int(manifest_id)),
                            branch=branch,
                            size=maybe_get(manifest_info, "size"),
                            download_size=maybe_get(manifest_info, "download"),
                        )
                    else:
                        encrypted_id = PrivateManifestInfo._get_id(depot, branch)
                        if encrypted_id is not None:
                            manifest = PrivateManifestInfo(state, encrypted_id, branch)
                        else:  # fall back to the public version
                            manifest = ManifestInfo(
                                state,
                                public_id,
                                branch=self.public_branch,
                                size=maybe_get(public_manifest_info, "size"),
                                download_size=maybe_get(public_manifest_info, "download"),
                            )

                    depot_ = Depot(
                        id=id,
                        name=name,
                        config=config,
                        app=self,
                        max_size=max_size,
                        shared_install=shared_install,
                        system_defined=system_defined,
                        branch=branch,
                        manifest=manifest,
                    )
                    manifest.depot = depot_
                    branch.depots.append(depot_)

    def get_branch(self, name: str) -> Branch | None:
        """Get a branch by its name."""
        return self._branches.get(name)

    @property
    def branches(self) -> Sequence[Branch]:
        """The branches for this app."""
        return list(self._branches.values())

    @property
    def public_branch(self) -> Branch:
        """The public branch. Shorthand for:

        .. code:: python

            app.get_branch("public")
        """
        return self._branches["public"]

    async def depots(self) -> Sequence[Depot | HeadlessDepot]:
        return [depot for branch in self._branches.values() for depot in branch.depots] + self.headless_depots  # type: ignore

    def is_on_windows(self) -> bool:
        """Whether the app is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the app is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the app is playable on Linux."""
        return self._on_linux

    def has_visible_stats(self) -> bool:
        """Whether the app has publicly visible stats."""
        return self._stats_visible

    def is_free(self) -> bool:
        """Whether the app is free to download."""
        return self._free

    def __repr__(self) -> str:
        attrs = ("name", "id", "type", "sha", "change_number")
        resolved = [f"{name}={getattr(self, name)!r}" for name in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"


class PackageInfo(ProductInfo, PartialPackage):
    """Represents a collection of information on a package."""

    __slots__ = (
        "sha",
        "size",
        "change_number",
        "_apps",
        "billing_type",
        "license_type",
        "status",
        "depot_ids",
        "app_items",
        "_language",
    )

    def __init__(
        self,
        state: ConnectionState,
        data: manifest.PackageInfo,
        proto: app_info.CMsgClientPicsProductInfoResponsePackageInfo,
    ):
        super().__init__(state, proto, id=proto.packageid)
        self._apps = [PartialApp(state, id=id) for id in data["appids"].values()]
        self.billing_type = BillingType.try_value(data["billingtype"])
        """The billing type for the package."""
        self.license_type = LicenseType.try_value(data["licensetype"])
        """The license type for the package."""
        self.status = PackageStatus.try_value(data["status"])
        """The status for the package."""
        self.depot_ids = cast("list[DepotID]", list(data["depotids"].values()))
        """The depot IDs in of the depots in the package."""
        self.app_items = list(data["appitems"].values())
        """The app items in the package."""

    def __repr__(self) -> str:
        attrs = ("id", "sha", "change_number")
        resolved = [f"{name}={getattr(self, name)!r}" for name in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    @overload
    async def apps(self, *, language: Language) -> list[PartialApp[str]]: ...

    @overload
    async def apps(self, *, language: None = ...) -> list[PartialApp[None]]: ...

    async def apps(  # type: ignore[reportIncompatibleMethodOverride]
        self,
        *,
        language: Language | None = None,
    ) -> list[PartialApp[None]] | list[PartialApp[str]]:
        if language is not None:
            return await super().apps(language=language)
        return self._apps



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import weakref
from contextlib import asynccontextmanager
from datetime import datetime, timedelta
from functools import partial
from ipaddress import IPv4Address
from operator import attrgetter
from typing import TYPE_CHECKING, Any, Literal, cast

from ._const import DOCS_BUILDING, MISSING, UNIX_EPOCH, URL, TaskGroup
from .abc import BaseUser, Messageable
from .app import PartialApp
from .enums import Language, PersonaState, PersonaStateFlag, Type
from .id import _ID64_TO_ID32, ID
from .profile import ClientUserProfile, OwnedProfileItems, ProfileInfo, ProfileItem
from .protobufs import friend_messages, player
from .protobufs.friends import CMsgClientPersonaStateFriend as UserProto
from .reaction import Emoticon, MessageReaction, Sticker
from .trade import Asset, Inventory, Item, TradeOffer
from .types.id import ID32, AppID, ContextID
from .utils import DateTime, cached_slot_property, parse_bb_code

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Sequence

    from typing_extensions import Self

    from .app import App
    from .channel import UserChannel
    from .friend import Friend
    from .media import Media
    from .message import UserMessage
    from .state import ConnectionState
    from .types import user

__all__ = (
    "User",
    "ClientUser",
    "AnonymousClientUser",
)


class _BaseUser(BaseUser):
    __slots__ = (
        "name",
        "app",
        "state",
        "flags",
        "last_seen_online",
        "last_logoff",
        "last_logon",
        "rich_presence",
        "game_server_ip",
        "game_server_port",
        "_state",
        "_avatar_sha",
    )

    def __init__(self, state: ConnectionState, proto: UserProto):
        super().__init__(state, proto.friendid)
        self.rich_presence = None  # ensure the attr exists
        self._update(proto)

    def _update(self, proto: UserProto, /) -> None:
        self.name = proto.player_name
        """The user's username."""
        self._avatar_sha = proto.avatar_hash

        self.game_server_ip = IPv4Address(proto.game_server_ip) if proto.game_server_ip else None
        """The IP address of the game server the user is currently playing on."""
        self.game_server_port = proto.game_server_port or None
        """The port of the game server the user is currently playing on."""

        self.last_logoff = DateTime.from_timestamp(proto.last_logoff)
        """The last time the user logged off from steam."""
        self.last_logon = DateTime.from_timestamp(proto.last_logon)
        """The last time the user logged into steam."""
        self.last_seen_online = DateTime.from_timestamp(proto.last_seen_online)
        """The last time the user could be seen online."""
        if proto.is_set("rich_presence"):
            self.rich_presence = {message.key: message.value for message in proto.rich_presence}
        """The rich presence of the user."""
        self.app = (
            PartialApp(self._state, name=proto.game_name, id=proto.game_played_app_id)
            if proto.game_played_app_id
            else None
        )
        """The app the user is playing. Is ``None`` if the user isn't in a app or one that is recognised by the API."""
        self.state = PersonaState.try_value(proto.persona_state)
        """The current persona state of the account (e.g. LookingToTrade)."""
        self.flags = PersonaStateFlag.try_value(proto.persona_state_flags)
        """The persona state flags of the account."""


class User(_BaseUser, Messageable["UserMessage"]):
    """Represents a Steam user's account.

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: str(x)

            Returns the user's name.
    """

    __slots__ = ("_cs_channel",)

    @staticmethod
    def _dict_to_proto(data: user.User) -> UserProto:
        return UserProto(
            player_name=data["personaname"],
            friendid=int(data["steamid"]),
            avatar_hash=bytes.fromhex(data["avatarhash"]) if "avatarhash" in data else b"\x00" * 20,
            persona_state=data.get("personastate", 0),
            game_played_app_id=int(data.get("gameid", 0)),
            persona_state_flags=data.get("personastateflags", 0),
            last_logoff=data.get("lastlogoff", 0),
            game_name=data.get("gameextrainfo", 0),
        )

    async def add(self) -> None:
        """Sends a friend invite to the user to your friends list."""
        await self._state.add_user(self.id64)

    async def remove(self) -> None:
        """Remove the user from your friends list."""
        await self._state.remove_user(self.id64)
        self._state.user._friends.pop(self.id, None)

    async def cancel_invite(self) -> None:
        """Cancels an invitation sent to the user. This effectively does the same thing as :meth:`remove`."""
        await self._state.remove_user(self.id64)

    async def block(self) -> None:
        """Blocks the user."""
        await self._state.block_user(self.id64)

    async def unblock(self) -> None:
        """Unblocks the user."""
        await self._state.unblock_user(self.id64)

    async def escrow(self, token: str | None = None) -> timedelta | None:
        """Check how long any received items would take to arrive. ``None`` if the user has no escrow (hold) or has a
        private inventory.

        Parameters
        ----------
        token
            The user's trade offer token, not required if you are friends with the user.
        """
        data = await self._state.http.get_user_escrow(self.id64, token)
        if (their_escrow := data.get("their_escrow")) is None:  # private
            return None
        seconds = their_escrow["escrow_end_duration_seconds"]
        return timedelta(seconds=seconds) if seconds else None

    async def send(
        self,
        content: Any = MISSING,
        /,
        *,
        trade: TradeOffer[Asset[User], Asset[ClientUser], Any] | None = None,
        media: Media | None = None,
    ) -> UserMessage[ClientUser] | None:
        """Send a message, trade or image to an :class:`User`.

        Parameters
        ----------
        content
            The message to send to the user.
        trade
            The trade offer to send to the user.

            Note
            ----
            This will have its :attr:`~steam.TradeOffer.id` attribute updated after being sent.

        media
            The media to send to the user.

        Raises
        ------
        :exc:`~steam.HTTPException`
            Sending the message failed.
        :exc:`~steam.Forbidden`
            You do not have permission to send the message.

        Returns
        -------
        The sent message only applicable if ``content`` is passed.
        """
        super_send = partial(
            Messageable["UserMessage[ClientUser]"].send, cast("Messageable[UserMessage[ClientUser]]", self)
        )  # done because calls to super() break in WrapsUser subclasses

        message = await super_send(content, media=media)
        if trade is not None:
            await self._send_trade(trade)

        return message

    @cached_slot_property("_cs_channel")
    def _channel(self) -> UserChannel:
        from .channel import UserChannel

        return UserChannel(self._state, self)

    async def history(
        self, *, limit: int | None = 100, before: datetime | None = None, after: datetime | None = None
    ) -> AsyncGenerator[UserMessage, None]:
        from .message import UserMessage

        after = after or UNIX_EPOCH
        before = before or DateTime.now()
        after_timestamp = int(after.timestamp())
        before_timestamp = int(before.timestamp())
        yielded = 0

        last_message_timestamp = before_timestamp
        ordinal = 0

        while True:
            resp = await self._state.fetch_user_history(
                self.id64, start=after_timestamp, last=last_message_timestamp, start_ordinal=ordinal
            )

            message: friend_messages.GetRecentMessagesResponseFriendMessage | None = None

            for message in resp.messages:
                new_message = UserMessage._from_history(self._channel, message)
                new_message.created_at = DateTime.from_timestamp(message.timestamp)
                if not after < new_message.created_at < before:
                    return
                if limit is not None and yielded >= limit:
                    return

                new_message.author = self if message.accountid == self.id else self._state.user
                emoticon_reactions = [
                    MessageReaction(
                        self._state,
                        new_message,
                        Emoticon(self._state, r.reaction),
                        None,
                        self if reactor == self.id else self._state.user,
                    )
                    for r in message.reactions
                    if r.reaction_type == Emoticon._TYPE
                    for reactor in r.reactors
                ]
                sticker_reactions = [
                    MessageReaction(
                        self._state,
                        new_message,
                        None,
                        Sticker(self._state, r.reaction),
                        self if reactor == self.id else self._state.user,
                    )
                    for r in message.reactions
                    if r.reaction_type == Sticker._TYPE
                    for reactor in r.reactors
                ]
                new_message.reactions = emoticon_reactions + sticker_reactions

                yield new_message
                yielded += 1

            if message is None:
                return

            last_message_timestamp = message.timestamp
            ordinal = message.ordinal

            if not resp.more_available:
                return

    @asynccontextmanager
    async def typing(self) -> AsyncGenerator[None, None]:
        """Send a typing indicator continuously to the channel while in the context manager.

        Note
        ----
        This only works in DMs.

        Usage:

        .. code:: python

            async with channel.typing():
                ...  # do your expensive operations
        """

        async def inner() -> None:
            while True:
                await self.trigger_typing()
                await asyncio.sleep(10)

        async with TaskGroup() as tg:
            t = tg.create_task(inner())
            yield
            t.cancel()

    async def trigger_typing(self) -> None:
        """Send a typing indicator to the channel once.

        Note
        ----
        This only works in DMs.
        """
        await self._state.send_user_typing(self.id64)


class ClientUser(_BaseUser):
    """Represents your account.

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: str(x)

            Returns the user's name.
    """

    # TODO more stuff to add https://github.com/DoctorMcKay/node-steamcommunity/blob/master/components/profile.js

    __slots__ = ("_friends", "_inventory_locks")

    def __init__(self, state: ConnectionState, proto: UserProto):
        super().__init__(state, proto)
        self._friends: dict[ID32, Friend] = {}
        self._inventory_locks = weakref.WeakValueDictionary[AppID, asyncio.Lock]()

    async def friends(self) -> Sequence[Friend]:
        """A list of the user's friends."""
        return list(self._friends.values())

    def get_friend(self, id: int, /) -> Friend | None:
        """Get a friend from the client user's friends list.

        Parameters
        ----------
        id
            The ID64 of the friend to get.
        """
        return self._friends.get(_ID64_TO_ID32(id))

    async def inventory(
        self, app: App, /, *, context_id: int | None = None, language: Language | None = None
    ) -> Inventory[Item[Self], Self]:
        try:
            lock = self._inventory_locks[app.id]
        except KeyError:
            lock = self._inventory_locks[app.id] = asyncio.Lock()

        if context_id is None:
            context_id = 6 if app.name == "Steam" and context_id is None else 2
        context_id = ContextID(context_id)
        async with lock:  # requires a per-app lock to avoid Result.DuplicateRequest
            resp = await self._state.fetch_user_inventory(
                self.id64, app.id, context_id, language
            )  # fast path for own account cause it works for this
            return Inventory(self._state, resp, self, app, context_id, language)

    async def setup_profile(self) -> None:
        """Set up your profile if possible."""
        params = {"welcomed": 1}
        await self._state.http.get(URL.COMMUNITY / "my/edit", params=params)

    async def clear_nicks(self) -> None:
        """Clears the client user's nickname/alias history."""
        await self._state.http.clear_nickname_history()

    async def profile_items(self, *, language: Language | None = None) -> OwnedProfileItems[Self]:
        """Fetch all the client user's profile items.

        Parameters
        ----------
        language
            The language to fetch the profile items in. If ``None`` the current language is used
        """
        items = await self._state.fetch_profile_items(language)
        return OwnedProfileItems(
            backgrounds=[
                ProfileItem(self._state, self, background, um=player.SetProfileBackgroundRequest)
                for background in items.profile_backgrounds
            ],
            mini_profile_backgrounds=[
                ProfileItem(self._state, self, mini_profile_background, um=player.SetMiniProfileBackgroundRequest)
                for mini_profile_background in items.mini_profile_backgrounds
            ],
            avatar_frames=[
                ProfileItem(self._state, self, avatar_frame, um=player.SetAvatarFrameRequest)
                for avatar_frame in items.avatar_frames
            ],
            animated_avatars=[
                ProfileItem(self._state, self, animated_avatar, um=player.SetAnimatedAvatarRequest)
                for animated_avatar in items.animated_avatars
            ],
            modifiers=[ProfileItem(self._state, self, modifier) for modifier in items.profile_modifiers],
        )

    async def profile(self, *, language: Language | None = None) -> ClientUserProfile:
        return ClientUserProfile(
            *await asyncio.gather(
                self.equipped_profile_items(language=language),
                self.profile_info(),
                self.profile_customisation_info(),
                self.profile_items(language=language),
            )
        )

    async def profile_info(self: ClientUser | Friend) -> ProfileInfo:
        """The friend's profile info."""
        info = await self._state.fetch_friend_profile_info(self.id64)
        # why this is friend only I'm not really sure considering it's available through the API
        return ProfileInfo(
            created_at=DateTime.from_timestamp(info.time_created),
            real_name=info.real_name or None,
            city_name=info.city_name or None,
            state_name=info.state_name or None,
            country_name=info.country_name or None,
            headline=info.headline or None,
            summary=parse_bb_code(info.summary),
        )

    async def edit(
        self,
        *,
        name: str | None = None,
        real_name: str | None = None,
        url: str | None = None,
        summary: str | None = None,
        country: str | None = None,
        state: str | None = None,
        city: str | None = None,
        avatar: Media | None = None,
        # TODO privacy params, use ums
    ) -> None:
        """Edit the client user's profile. Any values that aren't set will use their defaults.

        Parameters
        ----------
        name
            The new name you wish to go by.
        real_name
            The real name you wish to go by.
        url
            The custom url ending/path you wish to use.
        summary
            The summary/description you wish to use.
        country
            The country you want to be from.
        state
            The state you want to be from.
        city
            The city you want to be from.
        avatar
            The avatar you wish to use.

            Note
            ----
            This needs to be at least 184px x 184px.

        Raises
        -------
        :exc:`~steam.HTTPException`
            Editing your profile failed.
        """
        if any((name, real_name, url, summary, country, state, city)):
            await self._state.http.edit_profile_info(name, real_name, url, summary, country, state, city)
        if avatar is not None:
            await self._state.http.update_avatar(avatar, "player_avatar_image", sId=self.id64)
        # TODO privacy stuff


IndividualID = ID[Literal[Type.Individual]]


class WrapsUser(User if TYPE_CHECKING or DOCS_BUILDING else BaseUser, Messageable["UserMessage"]):
    """Internal class used for creating a User subclass optimised for memory. Composes the original user and forwards
    all of its attributes.

    Similar concept to discord.py's Member except more generalised for the larger number situations Steam throws at us.
    Slightly different however in that isinstance(SubclassOfWrapsUser(), User) should pass.

    Note
    ----
    This class does not forward ClientUsers attribute's so things like Clan().me.clear_nicks() will fail.

    If DOCS_BUILDING is True then this class behaves like a normal User because we need to be able to access the
    doc-strings reliably and memory usage isn't a concern.
    """

    __slots__ = ("_user", "_cs_channel")

    def __init__(self, state: ConnectionState, user: User | ClientUser):
        IndividualID.__init__(self, user.id64, type=Type.Individual)
        self._user = user

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        user_dict = User.__dict__.copy()
        user_dict.pop("__annotations__", None)
        for name, function in set(user_dict.items()) - set(object.__dict__.items()):
            if not name.startswith("__") and name not in User.__slots__:
                setattr(cls, name, function)

        if not DOCS_BUILDING:
            for name in _BaseUser.__slots__:
                setattr(cls, name, property(attrgetter(f"_user.{name}")))  # TODO time this with a compiled property
                # probably wont be different than the above

        User.register(cls)


class AnonymousClientUser(ID[Literal[Type.AnonUser]]):
    __slots__ = ("_state",)

    def __init__(self, state: ConnectionState, id64: int):
        super().__init__(id64, type=Type.AnonUser)
        self._state = state

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} universe={self.universe!r} instance={self.instance!r}>"



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: service_usernews.proto
# plugin: python-betterproto
# Last updated 09/09/2021
from dataclasses import dataclass

import betterproto

from .msg import UnifiedMessage


class GetAppDetailsSpotlightRequest(UnifiedMessage, um_name="UserNews.GetAppDetailsSpotlight"):
    appid: int = betterproto.uint32_field(1)
    include_already_seen: bool = betterproto.bool_field(2)


class GetAppDetailsSpotlightResponse(UnifiedMessage, um_name="UserNews.GetAppDetailsSpotlight"):
    @dataclass(eq=False, repr=False)
    class FeaturedEvent(betterproto.Message):
        event_type: int = betterproto.uint32_field(1)
        event_time: int = betterproto.uint32_field(2)
        clan_id: int = betterproto.fixed64_field(3)
        clan_announcementid: int = betterproto.fixed64_field(4)
        appid: int = betterproto.uint32_field(5)
        rtime32_last_modified: int = betterproto.uint32_field(6)

    events: list[FeaturedEvent] = betterproto.message_field(1)


class GetUserNewsRequest(UnifiedMessage, um_name="UserNews.GetUserNews"):
    count: int = betterproto.uint32_field(1)
    starttime: int = betterproto.uint32_field(2)
    endtime: int = betterproto.uint32_field(3)
    language: str = betterproto.string_field(4)
    filterflags: int = betterproto.uint32_field(5)
    filterappid: int = betterproto.uint32_field(6)


class GetUserNewsResponse(UnifiedMessage, um_name="UserNews.GetUserNews"):
    @dataclass(eq=False, repr=False)
    class Event(betterproto.Message):
        eventtype: int = betterproto.uint32_field(1)
        eventtime: int = betterproto.uint32_field(2)
        steamid_actor: int = betterproto.fixed64_field(3)
        steamid_target: int = betterproto.fixed64_field(4)
        gameid: int = betterproto.fixed64_field(5)
        packageid: int = betterproto.uint32_field(6)
        shortcutid: int = betterproto.uint32_field(7)
        achievement_names: list[str] = betterproto.string_field(8)
        clan_eventid: int = betterproto.fixed64_field(9)
        clan_announcementid: int = betterproto.fixed64_field(10)
        publishedfileid: int = betterproto.fixed64_field(11)
        event_last_mod_time: int = betterproto.uint32_field(12)
        appids: list[int] = betterproto.uint32_field(13)

    news: list[Event] = betterproto.message_field(1)

    @dataclass(eq=False, repr=False)
    class AchievementDisplayData(betterproto.Message):
        appid: int = betterproto.uint32_field(1)

        @dataclass(eq=False, repr=False)
        class Achievement(betterproto.Message):
            name: str = betterproto.string_field(1)
            display_name: str = betterproto.string_field(2)
            display_description: str = betterproto.string_field(3)
            icon: str = betterproto.string_field(4)
            unlocked_pct: float = betterproto.float_field(5)
            hidden: bool = betterproto.bool_field(6)

        achievements: list[Achievement] = betterproto.message_field(2)

    achievement_display_data: list[AchievementDisplayData] = betterproto.message_field(2)



import betterproto

from .base import CClanEventData
from .msg import UnifiedMessage


class RespondToClanInviteRequest(UnifiedMessage, um_name="Clan.RespondToClanInvite"):
    steamid: int = betterproto.fixed64_field(1)
    accept: bool = betterproto.bool_field(2)


class RespondToClanInviteResponse(UnifiedMessage, um_name="Clan.RespondToClanInvite"):
    pass


# not public and also non-functional atm


class GetAdjacentPartnerEventsRequest(UnifiedMessage, um_name="Clan.GetAdjacentPartnerEvents"):
    steam_id: int = betterproto.fixed64_field(1)  # might be account id
    announcement_id: int = betterproto.fixed64_field(2)
    # count_later: int = betterproto.uint32_field(2)  # might allow announcement_id
    # might allow the limit


class GetAdjacentPartnerEventsResponse(UnifiedMessage, um_name="Clan.GetAdjacentPartnerEvents"):
    events: list[CClanEventData] = betterproto.message_field(1)


class GetEventDetailsRequest(UnifiedMessage, um_name="Clan.GetEventDetails"):
    event_ids: list[int] = betterproto.fixed64_field(1)
    clan_ids: list[int] = betterproto.fixed32_field(2)


class GetEventDetailsResponse(UnifiedMessage, um_name="Clan.GetEventDetails"):
    events: list[CClanEventData] = betterproto.message_field(1)


class GetSinglePartnerEventRequest(UnifiedMessage, um_name="Clan.GetSinglePartnerEvent"):
    clan_id: int = betterproto.fixed32_field(1)
    announcement_id: int = betterproto.uint64_field(2)


class GetSinglePartnerEventResponse(UnifiedMessage, um_name="Clan.GetSinglePartnerEvent"):
    events: list[CClanEventData] = betterproto.message_field(1)



from __future__ import annotations

import asyncio
from pathlib import Path
from typing import TYPE_CHECKING, Final, Literal, cast, overload

from typing_extensions import Self

from ..._const import DOCS_BUILDING, VDF_LOADS, timeout
from ..._gc import Client as Client_
from ..._gc.client import ClientUser as ClientUser_
from ...app import TF2, App
from ...ext import commands
from ...utils import cached_property  # noqa: TC001
from .protobufs import struct_messages
from .state import GCState  # noqa: TC001

if TYPE_CHECKING:
    import os
    from collections.abc import Collection

    from ...enums import Language as Language_
    from ...ext import tf2
    from ...trade import Inventory, Item
    from .backpack import Backpack, BackpackItem, Schema

__all__ = (
    "Client",
    "Bot",
)


class ClientUser(ClientUser_):
    if TYPE_CHECKING:

        @overload
        async def inventory(self, app: Literal[TF2], *, language: object = ...) -> Backpack:  # type: ignore
            ...

        @overload
        async def inventory(self, app: App, *, language: Language_ | None = None) -> Inventory[Item[Self], Self]:  # type: ignore
            ...


class Client(Client_):
    _APP: Final = TF2
    _ClientUserCls = ClientUser
    _state: GCState  # type: ignore  # PEP 705
    if TYPE_CHECKING:

        @cached_property
        def user(self) -> ClientUser: ...

    @property
    def schema(self) -> Schema:
        """TF2's item schema. ``None`` if the user isn't ready."""
        return self._state.schema

    @property
    def backpack_slots(self) -> int:
        """The client's number of backpack slots."""
        if self._state.backpack_slots is None:
            raise RuntimeError("GC isn't ready yet")
        return self._state.backpack_slots

    def is_premium(self) -> bool:
        """Whether or not the client's account has TF2 premium. ``None`` if the user isn't ready."""
        return self._state._is_premium  # type: ignore

    def set_language(self, file: os.PathLike[str]) -> None:  # TODO this doesn't work
        """Set the localization files for your bot.

        This isn't necessary in most situations.
        """
        file = Path(file).resolve()
        self._state.localisation = VDF_LOADS(file.read_text())

    async def craft(
        self, items: Collection[BackpackItem[ClientUser]], recipe: int = -2
    ) -> list[BackpackItem[ClientUser]]:
        """Craft a set of items together with an optional recipe.

        Parameters
        ----------
        items
            The items to craft.
        recipe
            The recipe to craft them with default is -2 (wildcard). Setting for metal crafts isn't required. See
            https://raw.githubusercontent.com/Gobot1234/TF2-Crafting-Recipe/master/craftRecipe.json for other recipe
            details.

        Returns
        -------
        The crafted items, ``None`` if crafting failed.
        """

        def check_craft(msg: struct_messages.CraftResponse) -> bool:
            if not msg.being_used:  # craft queue is FIFO, so this works fine
                msg.being_used = True
                return True

            return False

        future = self._state.ws.gc_wait_for(struct_messages.CraftResponse, check=check_craft)
        await self._state.ws.send_gc_message(
            struct_messages.CraftRequest(recipe=recipe, items=[item.id for item in items])
        )

        try:
            async with timeout(60):
                resp = await future
        except asyncio.TimeoutError:
            raise ValueError("crafting failed")
        else:
            if resp.recipe_id == -1:  # error occurred
                raise ValueError("crafting failed")

        return cast("list[BackpackItem[ClientUser]]", await asyncio.gather(*map(self._state.wait_for_item, resp.ids)))

    if TYPE_CHECKING or DOCS_BUILDING:

        async def on_account_update(self) -> None:
            """Called when the client user's account is updated. This can happen from any one of the below changing:

            - :meth:`is_premium`
            - :attr:`backpack_slots`
            """

        async def on_item_receive(self, item: tf2.BackpackItem) -> None:
            """Called when the client receives an item.

            Parameters
            ----------
            item
                The received item.
            """

        async def on_item_remove(self, item: tf2.BackpackItem) -> None:
            """Called when the client has an item removed from its backpack.

            Parameters
            ----------
            item
                The removed item.
            """

        async def on_item_update(self, before: tf2.BackpackItem, after: tf2.BackpackItem) -> None:
            """Called when the client has an item in its backpack updated.

            Parameters
            ----------
            before
                The item before being updated.
            after
                The item now.
            """


class Bot(commands.Bot, Client):
    pass



"""
The shared base classes for chat group interactions.

Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE
"""

from __future__ import annotations

import abc
import asyncio
import logging
from operator import attrgetter
from typing import TYPE_CHECKING, Any, Generic, Literal, Protocol, TypeAlias, cast, overload, runtime_checkable

from typing_extensions import Self, TypeVar, get_original_bases

from . import utils
from ._const import UNIX_EPOCH, _HasChatGroupMixin
from .abc import Channel, ClanT, GroupT, Message, PartialUser
from .app import PartialApp
from .enums import ChatMemberJoinState, ChatMemberRank, Type
from .errors import WSException
from .id import _ID64_TO_ID32, ID
from .models import Avatar
from .protobufs import chat
from .reaction import Emoticon, MessageReaction, PartialMessageReaction, Sticker
from .role import Role
from .types.id import ID32, ChatGroupID, ChatID, Intable, RoleID
from .types.user import IndividualID
from .user import ClientUser, User, WrapsUser
from .utils import DateTime, cached_slot_property

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable, Coroutine, Iterable, Sequence
    from datetime import datetime

    from .clan import Clan
    from .group import Group
    from .invite import ChatGroupInvite
    from .media import Media
    from .message import ClanMessage, GroupMessage
    from .state import ConnectionState


ChatT = TypeVar("ChatT", bound="Chat", default="Chat", covariant=True)
MemberT = TypeVar("MemberT", bound="Member", default="Member", covariant=True)
log = logging.getLogger(__name__)


@runtime_checkable
class _PartialMemberProto(
    _HasChatGroupMixin,
    Protocol,
    IndividualID if TYPE_CHECKING else object,  # type: ignore   # needs intersection types
):
    __slots__ = SLOTS = (
        "clan",
        "group",
        "rank",
        "join_state",
        "kick_expires_at",
        "_role_ids",
        "_state",
    )
    if not TYPE_CHECKING:
        __slots__ = ()

    _state: ConnectionState
    rank: ChatMemberRank
    join_state: ChatMemberJoinState
    _role_ids: tuple[RoleID, ...]
    kick_expires_at: datetime
    clan: Clan | None
    group: Group | None

    def _update(self, member: chat.Member, /) -> None:
        self.rank = ChatMemberRank.try_value(member.rank)
        self.join_state = ChatMemberJoinState.try_value(member.state)
        self._role_ids = cast("tuple[RoleID, ...]", tuple(member.role_ids))
        self.kick_expires_at = DateTime.from_timestamp(member.time_kick_expire)

    @property
    def roles(self) -> list[Role]:
        """The member's roles."""
        chat_group = self._chat_group
        return [chat_group._roles[role_id] for role_id in self._role_ids]

    @property
    def top_role(self) -> Role | None:
        """The member's top role."""
        return min(self.roles, key=attrgetter("ordinal"), default=None)

    async def add_role(self, role: Role, /) -> None:
        """Add a role to the member."""
        raise NotImplementedError
        chat_group = self._chat_group
        await chat_group.add_role(self, role)

    async def remove_role(self, role: Role, /) -> None:
        raise NotImplementedError
        chat_group = self._chat_group
        await chat_group.remove_role(self, role)

    async def mute(self, expires_at: datetime) -> None:
        """Mutes the member from the chat group."""
        chat_group = self._chat_group
        await chat_group.mute(self, expires_at)

    async def kick(self, expires_at: datetime) -> None:
        """Kicks the member from the chat group."""
        chat_group = self._chat_group
        await chat_group.kick(self, expires_at)

    async def ban(self) -> None:
        """Bans the member from the chat group."""
        chat_group = self._chat_group
        await chat_group.ban(self)


class PartialMember(PartialUser, _PartialMemberProto):  # type: ignore
    __slots__ = _PartialMemberProto.SLOTS

    clan: Clan | None
    group: Group | None

    @overload
    def __init__(self, state: ConnectionState, *, clan: Clan, member: chat.Member) -> None: ...

    @overload
    def __init__(self, state: ConnectionState, *, group: Group, member: chat.Member) -> None: ...

    def __init__(
        self, state: ConnectionState, *, clan: Clan | None = None, group: Group | None = None, member: chat.Member
    ) -> None:
        super().__init__(state, member.accountid)
        self.clan = clan
        self.group = group
        self._state = state
        self._update(member)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id!r} {f'clan={self.clan!r}' if self.clan else f'group={self.group!r}'} rank={self.rank!r}>"

    def copy(self) -> Self:
        return self.__class__(
            self._state,
            self.clan,
            self.group,
            chat.Member(
                self.id,
                self.join_state,  # type: ignore
                self.rank,  # type: ignore
                int(self.kick_expires_at.timestamp()),
                list(self._role_ids),
            ),
        )


if TYPE_CHECKING:

    class _BaseMember(WrapsUser, PartialMember, _PartialMemberProto):  # type: ignore
        pass

else:
    _BaseMember = WrapsUser


@PartialMember.register
class Member(_BaseMember, _HasChatGroupMixin, Generic[ClanT, GroupT]):
    """Represents a member of a chat group."""

    __slots__ = tuple(slot for slot in _BaseMember.__slots__ if slot not in {"_state"})

    def __init__(
        self, state: ConnectionState, chat_group: ChatGroup[Any, Any], user: User | ClientUser, member: chat.Member
    ) -> None:
        super().__init__(state, user)
        self.clan = cast(ClanT, None)
        self.group = cast(GroupT, None)

        self._update(member)

    _update: Callable[[chat.Member], None] = PartialMember._update  # type: ignore

    if not TYPE_CHECKING:
        roles = PartialMember.roles

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id!r} {f'clan={self.clan!r}' if self.clan else f'group={self.group!r}'} rank={self.rank!r}>"

    def copy(self) -> Self:
        return self.__class__(
            self._state,
            self._chat_group,
            self._user,
            chat.Member(
                self.id,
                chat.EChatRoomJoinState.Joined,
                self.rank,  # type: ignore
                int(self.kick_expires_at.timestamp()),
                list(self._role_ids),
            ),
        )


AuthorT = TypeVar("AuthorT", bound="PartialMember", default="PartialMember | Member", covariant=True)


class ChatMessage(Message[AuthorT, ChatT], Generic[AuthorT, MemberT, ChatT]):
    channel: ChatT
    author: AuthorT

    def __init__(self, proto: chat.IncomingChatMessageNotification, channel: ChatT, author: AuthorT) -> None:
        super().__init__(channel, proto)
        self.author = author
        self.created_at = DateTime.from_timestamp(proto.timestamp)
        self._mentions_ids = cast(tuple[ID32, ...], tuple(proto.mentions.ids))
        self.mentions_all = proto.mentions.mention_all
        self.mentions_here = proto.mentions.mention_here

    @classmethod
    def _from_history(
        cls, channel: ChatT, proto: chat.GetMessageHistoryResponseChatMessage  # type: ignore  # this is a constructor covariance is fine
    ) -> Self:
        self = cls.__new__(cls)  # skip __init__
        super().__init__(self, channel, proto)
        # if you want a fun time try and figure out the issues in calling these methods
        self.created_at = DateTime.from_timestamp(proto.server_timestamp)
        return self

    @property
    def _chat_group(self) -> Clan | Group:
        return self.channel._chat_group

    @property
    def mentions(self) -> list[MemberT | PartialMember]:
        """The members mentioned in the message."""
        return self._chat_group._maybe_members(self._mentions_ids)

    async def delete(self) -> None:
        """Deletes the message."""
        await self._state.delete_chat_messages(
            *self.channel._location, (int(self.created_at.timestamp()), self.ordinal)
        )

    async def fetch_reaction(self, emoticon: Emoticon | Sticker, /) -> list[MessageReaction]:
        """Fetches the reactions to this message with a given emoticon."""
        reactors = await self._state.fetch_message_reactors(
            *self.channel._location,
            server_timestamp=int(self.created_at.timestamp()),
            ordinal=self.ordinal,
            reaction_name=str(emoticon) if isinstance(emoticon, Emoticon) else emoticon.name,
            reaction_type=emoticon._TYPE,
        )

        reaction = utils.find(
            lambda r: (r.emoticon or r.sticker).name == emoticon.name, self.partial_reactions  # type: ignore
        )
        assert reaction
        return [
            MessageReaction(
                self._state,
                self,
                reaction.emoticon,  # type: ignore
                reaction.sticker,  # type: ignore  # needs conditional types
                user=self._chat_group._maybe_member(reactor),
            )
            for reactor in reactors
        ]

    async def fetch_reactions(self) -> list[MessageReaction]:
        """Fetches all this message's reactions.

        Warning
        -------
        This does nothing for messages that aren't fetched by :meth:`GroupChannel.history`/:meth:`ClanChannel.history`.
        """
        return [
            r
            for rs in await asyncio.gather(
                *(
                    self.fetch_reaction(reaction.emoticon or reaction.sticker)  # type: ignore
                    for reaction in self.partial_reactions
                )
            )
            for r in rs
        ]

    async def _react(self, emoticon: Sticker | Emoticon, add: bool) -> None:
        await self._state.react_to_chat_message(
            *self.channel._location,
            int(self.created_at.timestamp()),
            self.ordinal,
            str(emoticon) if isinstance(emoticon, Emoticon) else emoticon.name,
            reaction_type=emoticon._TYPE,
            is_add=add,
        )
        reaction = MessageReaction(
            self._state,
            self,
            emoticon if isinstance(emoticon, Emoticon) else None,  # type: ignore
            emoticon if isinstance(emoticon, Sticker) else None,  # type: ignore
            self._state.user,
            DateTime.now(),
            self.ordinal,
        )

        if add:
            self.reactions.append(reaction)
        else:
            try:
                self.reactions.remove(reaction)
            except ValueError:
                log.debug("Reaction removed for a reaction that wasn't added")
        self._state.dispatch(f"reaction_{'add' if add else 'remove'}", reaction)

    async def ack(self) -> None:
        await self._state.ack_chat_message(*self.channel._location, int(self.created_at.timestamp()))


ChatMessageT = TypeVar(
    "ChatMessageT", bound="GroupMessage | ClanMessage", default="GroupMessage | ClanMessage", covariant=True
)

GroupChannelProtos: TypeAlias = chat.IncomingChatMessageNotification | chat.State | chat.ChatRoomState


class Chat(Channel[ChatMessageT, ClanT, GroupT], _HasChatGroupMixin):
    __slots__ = ("id", "name", "joined_at", "position", "last_message", "_cs_location")

    def __init__(
        self, state: ConnectionState, chat_group: ChatGroup[Any, Self], proto: GroupChannelProtos
    ):  # chat_group is purposely unused for type checking purposes with chat_cls
        super().__init__(state)
        self.id = ChatID(proto.chat_id)
        """The ID of the channel."""
        self.name: str | None = None
        """
        The name of the channel, this could be the same as the :attr:`~steam.Clan.name`/:attr:`~steam.Group.name`,
        if it's the main channel.
        """
        self.joined_at: datetime | None = None
        """The time the client joined the chat."""
        self.position: int | None = None
        """The position of the channel in the channel list."""
        self.last_message: ChatMessageT | None = None
        """The last message sent in the channel."""
        self._update(proto)

    def _update(self, proto: GroupChannelProtos, /) -> None:
        if isinstance(proto, (chat.State, chat.IncomingChatMessageNotification)):
            first, _, second = proto.chat_name.partition(" | ")
            self.name = second or first

        self.joined_at = (
            DateTime.from_timestamp(proto.time_joined) if isinstance(proto, chat.ChatRoomState) else self.joined_at
        )
        self.position = proto.sort_order if isinstance(proto, chat.State) else self.position

    def __repr__(self) -> str:
        attrs = ("name", "id", "group" if self.group is not None else "clan", "position")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Chat) and self._location == other._location

    def __hash__(self) -> int:
        return hash(self._location)

    @cached_slot_property("_cs_location")
    def _location(self) -> tuple[ChatGroupID, ChatID]:
        chat_id = self._chat_group._id
        assert chat_id is not None
        return chat_id, self.id

    @utils.classproperty
    def _type_args(cls: type[Self]) -> tuple[type[ChatMessageT], type[ClanT], type[GroupT]]:  # type: ignore
        return get_original_bases(cls)[0].__args__

    def _send_message(self, content: str) -> Coroutine[Any, Any, ChatMessageT]:
        return self._state.send_chat_message(*self._location, content)  # type: ignore

    def _send_media(self, media: Media) -> Coroutine[Any, Any, None]:
        return self._state.http.send_chat_media(*self._location, media)

    async def history(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AsyncGenerator[ChatMessageT, None]:
        after = after or UNIX_EPOCH
        before = before or DateTime.now()

        after_timestamp = int(after.timestamp())
        before_timestamp = int(before.timestamp())
        last_message_timestamp = before_timestamp
        last_ordinal: int = getattr(self.last_message, "ordinal", 0)
        yielded = 0
        message_cls, *_ = self._type_args

        while True:
            resp = await self._state.fetch_chat_history(
                *self._location, start=after_timestamp, last=last_message_timestamp, last_ordinal=last_ordinal
            )
            message = None
            messages: list[ChatMessageT] = []

            for message in resp.messages:
                new_message = message_cls._from_history(self, message)  # type: ignore  # there's no easy way to do this
                if not after < new_message.created_at < before:
                    return
                if limit is not None and yielded >= limit:
                    return

                new_message.author = self._chat_group._maybe_member(_ID64_TO_ID32(message.sender))
                new_message.partial_reactions = [
                    PartialMessageReaction(
                        self._state,
                        new_message,
                        Emoticon(self._state, r.reaction) if r.reaction_type == Emoticon._TYPE else None,  # type: ignore
                        Sticker(self._state, r.reaction) if r.reaction_type == Sticker._TYPE else None,  # type: ignore
                    )
                    for r in message.reactions
                ]
                messages.append(new_message)
                yielded += 1

            if message is None:
                return

            last_message_timestamp = message.server_timestamp
            last_ordinal = message.ordinal

            for message_ in messages:
                yield message_

            if not resp.more_available:
                return

    async def edit(self, *, name: str | None = None, move_below: Self | None) -> None:
        """Edit the chat.

        Parameters
        ----------
        name
            The new name of the chat.
        move_below
            The chat to move this chat below.
        """
        await self._state.edit_chat(*self._location, name, move_below.id if move_below else None)

    async def delete(self) -> None:
        """Delete the chat."""
        await self._state.delete_chat(*self._location)

    async def bulk_delete(self, messages: Iterable[ChatMessage], /) -> None:
        """Bulk delete messages.

        Parameters
        ----------
        messages
            The messages to delete.
        """
        await self._state.delete_chat_messages(
            *self._location, *((int(m.created_at.timestamp()), m.ordinal) for m in messages)
        )

    async def fetch_message(self, id: int, /) -> ChatMessageT | None:
        message = await super().fetch_message(id)
        if message is not None:
            message.reactions = await message.fetch_reactions()
            return message


ChatGroupTypeT = TypeVar(
    "ChatGroupTypeT", bound=Literal[Type.Clan, Type.Chat], default=Literal[Type.Clan, Type.Chat], covariant=True
)


class ChatGroup(ID[ChatGroupTypeT], Generic[MemberT, ChatT, ChatGroupTypeT]):
    """Base class for :class:`steam.Clan` and :class:`steam.Group`."""

    __slots__ = (
        "name",
        "_id",
        "app",
        "tagline",
        "active_member_count",
        "chunked",
        "_state",
        "_members",
        "_partial_members",
        "_owner_id",
        "_mods",
        "_officers",
        "_top_members",
        "_channels",
        "_default_channel_id",
        "_roles",
        "_default_role_id",
        "_avatar_sha",
        "_invites",
    )
    name: str
    """The name of the chat group."""
    _id: ChatGroupID
    app: PartialApp | None
    """The :class:`steam.PartialApp` that the chat group is associated with."""
    tagline: str
    """The tagline of the chat group."""
    active_member_count: int
    """The number of users in the clan."""
    chunked: bool
    """Whether the chat group has been chunked."""
    _avatar_sha: bytes
    _state: ConnectionState

    _owner_id: ID32
    _top_members: list[ID32]

    _default_channel_id: ChatID
    _default_role_id: RoleID

    def __init__(self, state: ConnectionState, id: Intable, type: ChatGroupTypeT | None = None):
        super().__init__(id, type=type)
        self._state = state
        self._init()

    def _init(self) -> None:
        self.chunked = False
        self.app: PartialApp | None = None
        self._members: dict[ID32, MemberT] = {}
        self._partial_members: dict[ID32, chat.Member] = {}  # deleted after _members is populated (if chunked)
        self._channels: dict[ChatID, ChatT] = {}
        self._roles: dict[RoleID, Role] = {}
        self._officers: list[ID32] = []
        self._mods: list[ID32] = []
        self._invites: dict[ID32, ChatGroupInvite] = {}

    @classmethod
    async def _from_proto(
        cls,
        state: ConnectionState,
        proto: chat.GetChatRoomGroupSummaryResponse,
        *,
        id: ChatGroupID | None = None,
        maybe_chunk: bool = True,
    ) -> Self:
        self = cls(state, id if id is not None else proto.chat_group_id)
        self.name = proto.chat_group_name
        self._id = ChatGroupID(proto.chat_group_id)
        self.active_member_count = proto.active_member_count
        self._owner_id = ID32(proto.accountid_owner)
        self._top_members = cast(list[ID32], proto.top_members)
        self.tagline = proto.chat_group_tagline
        self.app = PartialApp(state, id=proto.appid) if proto.appid else self.app
        self._avatar_sha = proto.chat_group_avatar_sha

        self._default_role_id = RoleID(proto.default_role_id)
        self._update_channels(proto.chat_rooms, default_channel_id=proto.default_chat_id)

        if maybe_chunk:
            try:
                group_state = await self._state.set_chat_group_active(self._id)
            except WSException:
                pass
            else:
                self._update_group_state(group_state)

            await self._maybe_chunk()

        return self

    async def _maybe_chunk(self):
        if self._state.auto_chunk_chat_groups:
            await self.chunk()

        self._officers = [id for id, member in self._partial_members.items() if member.rank is ChatMemberRank.Officer]
        self._mods = [id for id, member in self._partial_members.items() if member.rank is ChatMemberRank.Moderator]

    @utils.classproperty
    def _type_args(cls: type[Self]) -> tuple[type[MemberT], type[ChatT], type[ChatGroupTypeT]]:  # type: ignore
        return get_original_bases(cls)[0].__args__

    async def _add_member(self, member: chat.Member) -> MemberT:
        member_cls, _, _ = self._type_args
        id32 = ID32(member.accountid)
        user = await self._state._maybe_user(id32)
        new_member = member_cls(self._state, self, user, member)
        self._members[id32] = new_member
        return new_member

    def _remove_member(self, member: chat.Member) -> MemberT | PartialMember | None:
        id32 = ID32(member.accountid)
        try:
            return self._members.pop(id32)
        except KeyError:
            return self._get_partial_member(id32)
        finally:
            self._partial_members.pop(id32, None)

    def _update_channels(
        self, channels: list[chat.State] | list[chat.ChatRoomState], *, default_channel_id: int | None = None
    ) -> None:
        _, chat_cls, _ = self._type_args
        for channel in channels:
            try:
                new_channel = self._channels[ChatID(channel.chat_id)]
            except KeyError:
                new_channel = chat_cls(self._state, self, channel)
                self._channels[new_channel.id] = new_channel
            else:
                new_channel._update(channel)

        if default_channel_id is not None:
            self._default_channel_id = ChatID(default_channel_id)

    def _update_header_state(self, proto: chat.GroupHeaderState, /) -> None:
        self.name = proto.chat_name
        self._owner_id = ID32(proto.accountid_owner)
        self.app = PartialApp(self._state, id=proto.appid)
        self.tagline = proto.tagline
        self._avatar_sha = proto.avatar_sha
        self._default_role_id = RoleID(proto.default_role_id) or self._default_role_id
        for role in proto.roles:
            for role_action in proto.role_actions:
                if role.role_id == role_action.role_id:
                    self._roles[role.role_id] = Role(self._state, self, role, role_action)  # type: ignore

    def _update_group_state(self, group_state: chat.GroupState):
        self._partial_members = cast(
            dict[ID32, chat.Member], {member.accountid: member for member in group_state.members}
        )
        self._roles = {
            RoleID(role.role_id): Role(self._state, self, role, permissions)
            for role in group_state.header_state.roles
            for permissions in group_state.header_state.role_actions
            if permissions.role_id == role.role_id
        }
        member_cls, _, _ = self._type_args
        try:
            self._members = {
                self._state.user.id: member_cls(
                    self._state,
                    self,
                    self._state.user,
                    self._partial_members[self._state.user.id],
                )
            }
        except KeyError:
            self._members = {}  # we aren't a member

    def __repr__(self) -> str:
        attrs = ("name", "id", "universe", "instance")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __str__(self) -> str:
        return self.name

    @property
    def members(self) -> Sequence[MemberT]:
        """A list of the chat group's members."""
        return list(self._members.values())

    def get_member(self, id64: int, /) -> MemberT | None:
        """Get a member of the chat group by id.

        Parameters
        ----------
        id64
            The 64 bit ID of the member to get.
        """
        return self._members.get(_ID64_TO_ID32(id64))

    @abc.abstractmethod
    def _get_partial_member(self, id: ID32, /) -> PartialMember:
        raise NotImplementedError

    def _maybe_member(self, id: ID32, /) -> MemberT | PartialMember:
        if (member := self._members.get(id)) is not None:
            return member

        if (user := self._state.get_user(id)) is not None:
            member_cls, _, _ = self._type_args
            self._members[id] = member = member_cls(self._state, self, user, self._partial_members.pop(id))
            return member

        return self._get_partial_member(id)

    def _maybe_members(self, ids: Iterable[ID32], /) -> list[MemberT | PartialMember]:
        members: list[MemberT | PartialMember] = []
        member_cls, _, _ = self._type_args
        for id in ids:
            if (member := self._members.get(id)) is not None:
                members.append(member)
            elif (user := self._state.get_user(id)) is not None:
                self._members[id] = member = member_cls(self._state, self, user, self._partial_members.pop(id))
                members.append(member)
            else:
                members.append(self._get_partial_member(id))

        return members

    @property
    def owner(self) -> MemberT | PartialMember:
        """The chat group's owner."""
        return self._maybe_member(self._owner_id)

    @property
    def officers(self) -> list[MemberT | PartialMember]:
        """A list of the chat groups's administrators."""
        return self._maybe_members(self._officers)

    @property
    def mods(self) -> list[MemberT | PartialMember]:
        """A list of the chat group's moderators."""
        return self._maybe_members(self._mods)

    @property
    def top_members(self) -> Sequence[MemberT | PartialMember]:
        """A list of the chat group's top members."""
        return self._maybe_members(self._top_members)

    @property
    def me(self) -> MemberT:
        """The client user's account in this chat group."""
        return self._members[self._state.user.id]

    @property
    def channels(self) -> Sequence[ChatT]:
        """A list of the chat group's channels."""
        return list(self._channels.values())

    def get_channel(self, id: int, /) -> ChatT | None:
        """Get a channel from cache.

        Parameters
        ----------
        id
            The ID of the channel.
        """
        return self._channels.get(ChatID(id))

    @property
    def default_channel(self) -> ChatT:
        """The group's default channel."""
        return self._channels[self._default_channel_id]

    @property
    def roles(self) -> Sequence[Role]:
        """A list of the group's roles."""
        return list(self._roles.values())

    def get_role(self, id: int, /) -> Role | None:
        """Get a role from cache.

        Parameters
        ----------
        id
            The ID of the role.
        """
        return self._roles.get(RoleID(id))

    @property
    def default_role(self) -> Role:
        """The group's default role."""
        return self._roles[self._default_role_id]

    @property
    def avatar(self) -> Avatar:
        """The chat group's avatar."""
        return Avatar(self._state, self._avatar_sha, suffix="256")

    @abc.abstractmethod
    async def chunk(self) -> Sequence[MemberT]:
        """Get a list of all members in the group."""
        self.chunked = True
        del self._partial_members
        return self.members

    async def search(self, name: str) -> Sequence[MemberT]:
        """Search for members in the chat group.

        Parameters
        ----------
        name
            The name of the member to search for.
        """
        msg: chat.SearchMembersResponse = await self._state.ws.send_um_and_wait(
            chat.SearchMembersRequest(
                chat_group_id=self._id,
                search_text=name,
            )
        )

        if self.chunked:
            return [self._members[ID32(user.accountid)] for user in msg.matching_members]

        return cast(
            list[MemberT],
            self._maybe_members(
                user.id for user in [self._state._store_user(user.persona) for user in msg.matching_members]
            ),
        )

    async def invite(self, user: User, /) -> None:
        """Invites a :class:`~steam.User` to the chat group.

        Parameters
        -----------
        user
            The user to invite to the chat group.
        """
        await self._state.invite_user_to_chat_group(user.id64, self._id)

    async def join(self, *, invite_code: str | None = None) -> None:
        """Joins the chat group.

        Parameters
        ----------
        invite_code
            The invite code to use to join the chat group.
        """
        await self._state.join_chat_group(self._id, invite_code)
        if self._state.auto_chunk_chat_groups:
            await self.chunk()

    async def leave(self) -> None:
        """Leaves the chat group."""
        await self._state.leave_chat_group(self._id)

    async def edit(self, *, name: str | None = None, tagline: str | None = None, avatar: Media | None = None) -> None:
        """Edits the chat group.

        Parameters
        ----------
        name
            The new name of the chat group. If ``None``, the name will not be changed.
        tagline
            The new tagline of the chat group. If ``None``, the tagline will not be changed.
        avatar
            The new avatar of the chat group. If ``None``, the avatar will not be changed.
        """
        await self._state.edit_chat_group(self._id, name, tagline, avatar)

    async def create_role(self, name: str) -> Role:
        """Creates a role in the chat group.

        Parameters
        ----------
        name
            The name of the role to create.
        """
        return await self._state.create_role(self._id, name)

    async def create_channel(self, name: str) -> ChatT:
        """Creates a channel in the chat group.

        Parameters
        ----------
        name
            The name of the channel to create.
        """
        state = await self._state.create_chat(self._id, name)
        _, chat_cls, _ = self._type_args
        return chat_cls(self._state, self, state)

    # async def create_invite(self) -> InviteLink:  # chat.CreateInviteLinkRequest
    #     ...
    #
    # async def fetch_invite(self, code: str) -> InviteLink:  # chat.GetChatInviteLinkResponse
    #     ...
    #
    # async def invites(self) -> list[InviteLink]:  # chat.GetChatInvitesResponse
    #     ...
    #
    async def mute(self, member: IndividualID, expires_at: datetime) -> None:
        """Mutes a member of the chat group.

        Parameters
        ----------
        member
            The member to mute.
        expires_at
            When the mute expires.
        """
        await self._state.mute_chat_group_member(self._id, member.id64, expires_at)

    async def kick(self, member: IndividualID, expires_at: datetime) -> None:
        """Kicks a member of the chat group.

        Parameters
        ----------
        member
            The member to kick.
        expires_at
            When the kick expires.
        """
        await self._state.kick_chat_group_member(self._id, member.id64, expires_at)

    async def ban(self, member: IndividualID) -> None:
        """Bans a member of the chat group."""
        await self._state.set_chat_group_ban_state(self._id, member.id64, True)

    async def unban(self, user: IndividualID) -> None:
        """Unbans a member of the chat group."""
        await self._state.set_chat_group_ban_state(self._id, user.id64, False)

    # async def bans(self) -> list[Ban]:  # chat.GetBanListRequest
    #     ...



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import base64
import hmac
from dataclasses import dataclass
from hashlib import sha1
from time import time
from typing import TYPE_CHECKING, Final, Literal, TypeAlias

from ._const import URL
from .errors import ConfirmationError

if TYPE_CHECKING:
    from .state import ConnectionState
    from .types.id import TradeOfferID


__all__ = (
    "get_authentication_code",
    "get_confirmation_code",
    "get_device_id",
    "Tags",
    "Confirmation",
)

AUTH_CODE_CHARS: Final = "23456789BCDFGHJKMNPQRTVWXY"
AUTH_CODE_CHARS_LEN: Final = len(AUTH_CODE_CHARS)


def _hmac(secret: str, buffer: bytes, /) -> bytes:
    return hmac.new(base64.b64decode(secret), buffer, digestmod=sha1).digest()


def get_authentication_code(shared_secret: str, timestamp: int | None = None) -> str:
    """Get a Steam Guard code for signing in.

    Parameters
    -----------
    shared_secret
        Base 64 encoded shared secret from Steam Guard.
    timestamp
        The Unix timestamp to generate the key for. Defaults to the timestamp returned by :func:`time.time`.
    """
    timestamp = timestamp if timestamp is not None else int(time())
    time_hmac = _hmac(shared_secret, (timestamp // 30).to_bytes(8, "big"))
    begin = time_hmac[19] & 0xF

    full_code = int.from_bytes(time_hmac[begin : begin + 4], "big") & 0x7FFFFFFF

    code: list[str] = []
    for _ in range(5):
        full_code, i = divmod(full_code, AUTH_CODE_CHARS_LEN)
        code.append(AUTH_CODE_CHARS[i])
    return "".join(code)


def get_confirmation_code(identity_secret: str, tag: str, timestamp: int | None = None) -> str:
    """Get a trade confirmation code.

    Parameters
    -----------
    identity_secret
        Base 64 encoded identity secret from Steam Guard.
    tag
        The confirmation tag to encode.
    timestamp
        The time to generate the key for. Defaults to the timestamp returned by :func:`time.time`.
    """
    timestamp = timestamp if timestamp is not None else int(time())
    buffer = timestamp.to_bytes(8, "big") + tag.encode("ascii")
    return base64.b64encode(_hmac(identity_secret, buffer)).decode()


def get_device_id(id64: int) -> str:
    """Get the device ID for a user's 64-bit ID.

    Parameters
    -----------
    id64
        The 64-bit Steam ID to generate the device ID for.
    """
    # it works, however it's different that one generated from mobile app

    hexed_steam_id = sha1(str(id64).encode("ascii")).hexdigest()
    partial_id = [
        hexed_steam_id[:8],
        hexed_steam_id[8:12],
        hexed_steam_id[12:16],
        hexed_steam_id[16:20],
        hexed_steam_id[20:32],
    ]
    return f'android:{"-".join(partial_id)}'


Tags: TypeAlias = Literal["conf", "details", "allow", "cancel", "list"]


@dataclass(repr=False, slots=True)
class Confirmation:
    _state: ConnectionState
    id: int
    nonce: int
    creator_id: TradeOfferID  # this isn't really always the trade ID, but for our purposes this is fine

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} creator_id={self.creator_id}>"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Confirmation) and self.creator_id == other.creator_id and self.id == other.id

    async def _confirm_params(self, tag: Tags) -> dict[str, str | int]:
        code, timestamp = await self._state._get_confirmation_code(tag)
        return {
            "p": self._state._device_id,
            "a": self._state.user.id64,
            "k": code,
            "t": timestamp,
            "m": "android",
            "tag": tag,
        }

    async def _perform_op(self, op: Tags, /) -> None:
        params = await self._confirm_params(op) | {"op": op, "cid": self.id, "ck": self.nonce}
        resp = await self._state.http.get(URL.COMMUNITY / "mobileconf/ajaxop", params=params)
        if not resp["success"]:
            raise ConfirmationError(resp.get("message", "Unknown error"))

    async def confirm(self) -> None:
        await self._perform_op("allow")

    async def cancel(self) -> None:
        await self._perform_op("cancel")



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import hashlib
import os
import struct
from time import time
from typing import TYPE_CHECKING, Any, Literal, Protocol, runtime_checkable

from .utils import cached_slot_property

__all__ = ("Media",)

if TYPE_CHECKING:
    from collections.abc import Callable

    from _typeshed import StrOrBytesPath
    from typing_extensions import Self


@runtime_checkable
class MediaIO(Protocol):
    def seekable(self) -> bool: ...

    def seek(self, offset: int, whence: int = ..., /) -> Any: ...

    def tell(self) -> int: ...

    def readable(self) -> bool: ...

    def read(self, size: int = ..., /) -> bytes: ...

    def close(self) -> Any: ...

    def fileno(self) -> int: ...


class Media:
    """A wrapper around common media files. Used for :meth:`steam.User.send`.

    Parameters
    ----------
    fp
        An media or path-like to pass to :func:`open`.
    spoiler
        Whether to mark the media as a spoiler.

    Note
    ----
    Currently supported media types include:
        - PNG
        - JPG/JPEG
        - GIF
        - WEBM
        - MPG/MPEG
        - MP4
        - OGV
    """

    __slots__ = ("fp", "spoiler", "name", "width", "height", "type", "_size_cs", "_tell")
    fp: MediaIO

    def __init__(self, fp: MediaIO | StrOrBytesPath | int, *, spoiler: bool = False):
        self.fp = fp if isinstance(fp, MediaIO) else open(fp, "rb")  # noqa
        if not (self.fp.seekable() and self.fp.readable()):
            raise ValueError(f"File buffer {fp!r} must be seekable and readable")

        self._tell = self.fp.tell()

        # from https://stackoverflow.com/questions/8032642
        headers = self.fp.read(36)
        match type := next(filter(None, (test(headers) for test in tests)), None):
            case "png":
                check = struct.unpack(">i", headers[4:8])[0]
                if check != 0x0D0A1A0A:
                    raise ValueError("Opened file's headers do not match a standard PNGs headers")
                width, height = struct.unpack(">ii", headers[16:24])
                type_prefix = "image"
            case "gif":
                width, height = struct.unpack("<HH", headers[6:10])
                type_prefix = "image"
            case "jpeg":
                try:
                    self.fp.seek(self._tell)  # read 0xff next
                    size = self._tell + 2
                    ftype = 0
                    while not 0xC0 <= ftype <= 0xCF or ftype in {0xC4, 0xC8, 0xCC}:
                        self.fp.seek(size, 1)
                        byte = self.fp.read(1)
                        while ord(byte) == 0xFF:
                            byte = self.fp.read(1)
                        ftype = ord(byte)
                        size = struct.unpack(">H", self.fp.read(2))[0] - 2
                    # we are at a SOFn block
                    self.fp.seek(1, 1)  # skip 'precision' byte.
                    height, width = struct.unpack(">HH", self.fp.read(4))
                    type_prefix = "image"
                except Exception as exc:
                    raise ValueError from exc
            case "webm" | "mp4" | "mpeg" | "ogv":
                width, height = 0, 0
                type_prefix = "video"
            case _:
                raise TypeError("Unsupported file format passed")
        self.type = f"{type_prefix}/{type}"
        self.spoiler = spoiler
        self.width = width
        self.height = height
        self.name = f'{int(time())}_{getattr(self.fp, "name", f"media.{type}")}'

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: object) -> None:
        self.fp.close()

    def read(self) -> bytes:
        self.fp.seek(self._tell)
        read = self.fp.read()
        self.fp.seek(self._tell)
        return read

    @cached_slot_property
    def size(self) -> int:
        try:
            size = os.stat(self.fp.fileno()).st_size - self._tell  # noqa: PTH116
        except OSError:  # slow fallback
            size = len(self.fp.read())
            self.fp.seek(self._tell)
        if size > 1024 * 1024 * 10:  # 10MiB
            raise ValueError("File is too large to upload")
        return size

    @staticmethod
    def hash(contents: bytes) -> str:
        return hashlib.sha1(contents).hexdigest()


tests: list[Callable[[bytes], str | None]] = []


@tests.append
def test_jpeg(h: bytes) -> Literal["jpeg"] | None:
    if h[6:10] in {b"JFIF", b"Exif"} or h[:3] == b"\xff\xd8\xff":
        return "jpeg"


@tests.append
def test_png(h: bytes) -> Literal["png"] | None:
    if h.startswith(b"\211PNG\r\n\032\n"):
        return "png"


@tests.append
def test_gif(h: bytes) -> Literal["gif"] | None:
    if h[:6] in {b"GIF87a", b"GIF89a"}:
        return "gif"


@tests.append
def test_webm(h: bytes) -> Literal["webm"] | None:
    if h[22:29] == b"\x82\x84webmB":
        return "webm"


@tests.append
def test_mp4(h: bytes) -> Literal["mp4"] | None:
    if h[4:8] == b"ftyp":
        return "mp4"


@tests.append
def test_mpeg(h: bytes) -> Literal["mpeg"] | None:
    if h[:3] == b"\x00\x00\x01":
        return "mpeg"


@tests.append
def test_ogv(h: bytes) -> Literal["ogv"] | None:
    if h[28:34] == b"video":
        return "ogv"



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_clientserver_login.proto
# plugin: python-betterproto
# Last updated 19/04/2022

from dataclasses import dataclass

import betterproto

from .msg import UnifiedMessage


@dataclass(eq=False, repr=False)
class DefinitionBadgeData(betterproto.Message):
    level: int = betterproto.int32_field(1)
    image: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CommunityItemData(betterproto.Message):
    item_name: str = betterproto.string_field(1)
    item_title: str = betterproto.string_field(2)
    item_description: str = betterproto.string_field(3)
    item_image_small: str = betterproto.string_field(4)
    item_image_large: str = betterproto.string_field(5)
    item_movie_webm: str = betterproto.string_field(6)
    item_movie_mp4: str = betterproto.string_field(7)
    animated: bool = betterproto.bool_field(8)
    badge_data: list[DefinitionBadgeData] = betterproto.message_field(9)
    item_movie_webm_small: str = betterproto.string_field(10)
    item_movie_mp4_small: str = betterproto.string_field(11)
    profile_theme_id: str = betterproto.string_field(12)


@dataclass(eq=False, repr=False)
class Definition(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    defid: int = betterproto.uint32_field(2)
    type: int = betterproto.int32_field(3)
    community_item_class: int = betterproto.int32_field(4)
    community_item_type: int = betterproto.uint32_field(5)
    point_cost: int = betterproto.int64_field(6)
    timestamp_created: int = betterproto.uint32_field(7)
    timestamp_updated: int = betterproto.uint32_field(8)
    timestamp_available: int = betterproto.uint32_field(9)
    quantity: int = betterproto.int64_field(10)
    internal_description: str = betterproto.string_field(11)
    active: bool = betterproto.bool_field(12)
    community_item_data: CommunityItemData = betterproto.message_field(13)
    timestamp_available_end: int = betterproto.uint32_field(14)
    bundle_defids: list[int] = betterproto.uint32_field(15)
    usable_duration: int = betterproto.uint32_field(16)
    bundle_discount: int = betterproto.uint32_field(17)


class AddReactionRequest(UnifiedMessage, um_name="LoyaltyRewards.AddReaction"):
    target_type: int = betterproto.int32_field(1)
    targetid: int = betterproto.uint64_field(2)
    reactionid: int = betterproto.uint32_field(3)


class AddReactionResponse(UnifiedMessage, um_name="LoyaltyRewards.AddReaction"):
    pass


class BatchedQueryRewardItemsRequest(UnifiedMessage, um_name="LoyaltyRewards.BatchedQueryRewardItems"):
    requests: "list[QueryRewardItemsRequest]" = betterproto.message_field(1)


class BatchedQueryRewardItemsResponse(UnifiedMessage, um_name="LoyaltyRewards.BatchedQueryRewardItems"):
    responses: "list[BatchedQueryRewardItemsResponseResponse]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class BatchedQueryRewardItemsResponseResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    response: "QueryRewardItemsResponse" = betterproto.message_field(2)


class GetActivePurchaseBonusesRequest(UnifiedMessage, um_name="LoyaltyRewards.GetActivePurchaseBonuses"):
    pass


class GetActivePurchaseBonusesResponse(UnifiedMessage, um_name="LoyaltyRewards.GetActivePurchaseBonuses"):
    bonuses: "list[PurchaseBonus]" = betterproto.message_field(1)


class GetEligibleAppsRequest(UnifiedMessage, um_name="LoyaltyRewards.GetEligibleApps"):
    pass


class GetEligibleAppsResponse(UnifiedMessage, um_name="LoyaltyRewards.GetEligibleApps"):
    apps: "list[GetEligibleAppsResponseEligibleApp]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetEligibleAppsResponseEligibleApp(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    has_items_anyone_can_purchase: bool = betterproto.bool_field(2)
    event_app: bool = betterproto.bool_field(3)
    hero_carousel_image: str = betterproto.string_field(4)


class GetPointsForSpendResponse(UnifiedMessage, um_name="LoyaltyRewards.GetPointsForSpend"):
    points: int = betterproto.int64_field(1)


class GetProfileCustomizationsConfigResponse(UnifiedMessage, um_name="LoyaltyRewards.GetProfileCustomizationsConfig"):
    points_cost: int = betterproto.uint32_field(1)
    upgrade_points_cost: int = betterproto.uint32_field(2)
    purchasable_customization_types: list[int] = betterproto.int32_field(3)
    upgradable_customization_types: list[int] = betterproto.int32_field(4)
    max_slots_per_type: int = betterproto.uint32_field(5)
    max_upgradable_level: int = betterproto.uint32_field(6)


class GetReactionConfigRequest(UnifiedMessage, um_name="LoyaltyRewards.GetReactionConfig"):
    pass


class GetReactionConfigResponse(UnifiedMessage, um_name="LoyaltyRewards.GetReactionConfig"):
    reactions: "list[GetReactionConfigResponseReactionConfig]" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class GetReactionConfigResponseReactionConfig(betterproto.Message):
    reactionid: int = betterproto.int32_field(1)
    points_cost: int = betterproto.uint32_field(2)
    points_transferred: int = betterproto.uint32_field(3)
    valid_target_types: list[int] = betterproto.int32_field(4)
    valid_ugc_types: list[int] = betterproto.uint32_field(5)


class GetReactionsRequest(UnifiedMessage, um_name="LoyaltyRewards.GetReactions"):
    target_type: int = betterproto.int32_field(1)
    targetid: int = betterproto.uint64_field(2)


class GetReactionsResponse(UnifiedMessage, um_name="LoyaltyRewards.GetReactions"):
    reactionids: list[int] = betterproto.uint32_field(1)


class GetReactionsSummaryForUserResponse(UnifiedMessage, um_name="LoyaltyRewards.GetReactionsSummaryForUser"):
    total: "list[GetReactionsSummaryForUserResponseBreakdown]" = betterproto.message_field(1)
    user_reviews: "list[GetReactionsSummaryForUserResponseBreakdown]" = betterproto.message_field(2)
    ugc: "list[GetReactionsSummaryForUserResponseBreakdown]" = betterproto.message_field(3)
    profile: "list[GetReactionsSummaryForUserResponseBreakdown]" = betterproto.message_field(4)
    forum_topics: "list[GetReactionsSummaryForUserResponseBreakdown]" = betterproto.message_field(5)
    comments: "list[GetReactionsSummaryForUserResponseBreakdown]" = betterproto.message_field(6)
    total_points_given: int = betterproto.int64_field(7)
    total_points_received: int = betterproto.int64_field(8)
    total_points_given: int = betterproto.int64_field(9)
    total_points_received: int = betterproto.int64_field(10)


@dataclass(eq=False, repr=False)
class GetReactionsSummaryForUserResponseBreakdown(betterproto.Message):
    reactionid: int = betterproto.int32_field(1)
    given: int = betterproto.uint32_field(2)
    received: int = betterproto.uint32_field(3)
    points_given: int = betterproto.int64_field(4)
    points_received: int = betterproto.int64_field(5)


class GetSummaryRequest(UnifiedMessage, um_name="LoyaltyRewards.GetSummary"):
    steamid: int = betterproto.fixed64_field(1)


class GetSummaryResponse(UnifiedMessage, um_name="LoyaltyRewards.GetSummary"):
    summary: "GetSummaryResponseSummary" = betterproto.message_field(1)
    timestamp_updated: int = betterproto.uint32_field(2)
    auditid_highwater: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class GetSummaryResponseSummary(betterproto.Message):
    points: int = betterproto.int64_field(1)
    points_earned: int = betterproto.int64_field(2)
    points_spent: int = betterproto.int64_field(3)


class QueryRewardItemsRequest(UnifiedMessage, um_name="LoyaltyRewards.QueryRewardItems"):
    appids: list[int] = betterproto.uint32_field(1)
    time_available: int = betterproto.uint32_field(2)
    community_item_classes: list[int] = betterproto.int32_field(3)
    language: str = betterproto.string_field(4)
    count: int = betterproto.int32_field(5)
    cursor: str = betterproto.string_field(6)
    sort: int = betterproto.int32_field(7)
    sort_descending: bool = betterproto.bool_field(8)
    reward_types: list[int] = betterproto.int32_field(9)
    excluded_community_item_classes: list[int] = betterproto.int32_field(10)
    definitionids: list[int] = betterproto.uint32_field(11)
    filters: list[int] = betterproto.int32_field(12)
    filter_match_all_category_tags: list[str] = betterproto.string_field(13)
    filter_match_any_category_tags: list[str] = betterproto.string_field(14)
    contains_definitionids: list[int] = betterproto.uint32_field(15)
    include_direct_purchase_disabled: bool = betterproto.bool_field(16)
    excluded_content_descriptors: list[int] = betterproto.uint32_field(17)
    excluded_appids: list[int] = betterproto.uint32_field(18)


class QueryRewardItemsResponse(UnifiedMessage, um_name="LoyaltyRewards.QueryRewardItems"):
    definitions: "list[Definition]" = betterproto.message_field(1)
    total_count: int = betterproto.int32_field(2)
    count: int = betterproto.int32_field(3)
    next_cursor: str = betterproto.string_field(4)


class RedeemPointsRequest(UnifiedMessage, um_name="LoyaltyRewards.RedeemPoints"):
    defid: int = betterproto.uint32_field(1)
    expected_points_cost: int = betterproto.int64_field(2)


class RedeemPointsResponse(UnifiedMessage, um_name="LoyaltyRewards.RedeemPoints"):
    communityitemid: int = betterproto.uint64_field(1)
    bundle_community_item_ids: list[int] = betterproto.uint64_field(2)


class RedeemPointsForBadgeLevelRequest(UnifiedMessage, um_name="LoyaltyRewards.RedeemPointsForBadgeLevel"):
    defid: int = betterproto.uint32_field(1)
    num_levels: int = betterproto.int32_field(2)


class RedeemPointsForProfileCustomizationRequest(
    UnifiedMessage, um_name="LoyaltyRewards.RedeemPointsForProfileCustomization"
):
    customization_type: int = betterproto.int32_field(1)


class RedeemPointsForProfileCustomizationResponse(
    UnifiedMessage, um_name="LoyaltyRewards.RedeemPointsForProfileCustomization"
):
    purchaseid: int = betterproto.uint64_field(1)


class RedeemPointsForProfileCustomizationUpgradeRequest(
    UnifiedMessage, um_name="LoyaltyRewards.RedeemPointsForProfileCustomizationUpgrade"
):
    customization_type: int = betterproto.int32_field(1)
    new_level: int = betterproto.uint32_field(2)


class RedeemPointsForProfileCustomizationUpgradeResponse(
    UnifiedMessage, um_name="LoyaltyRewards.RedeemPointsForProfileCustomizationUpgrade"
):
    pass


class RegisterForSteamDeckRewardsRequest(UnifiedMessage, um_name="LoyaltyRewards.RegisterForSteamDeckRewards"):
    serial_number: str = betterproto.string_field(1)
    controller_code: str = betterproto.string_field(2)


class RegisterForSteamDeckRewardsResponse(UnifiedMessage, um_name="LoyaltyRewards.RegisterForSteamDeckRewards"):
    granted_profile_modifier: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class PurchaseBonus(betterproto.Message):
    bonusid: int = betterproto.uint64_field(1)
    appid: int = betterproto.uint32_field(2)
    active: bool = betterproto.bool_field(3)
    points: int = betterproto.int32_field(4)
    timestamp_start: int = betterproto.uint32_field(5)
    timestamp_end: int = betterproto.uint32_field(6)
    internal_description: str = betterproto.string_field(7)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: service_steamnotification.proto
# plugin: python-betterproto
# Last updated 05/11/2022

from dataclasses import dataclass

import betterproto

from .msg import UnifiedMessage


class GetPreferencesRequest(UnifiedMessage, um_name="SteamNotification.GetPreferences"):
    pass


class GetPreferencesResponse(UnifiedMessage, um_name="SteamNotification.GetPreferences"):
    preferences: "list[SteamNotificationPreference]" = betterproto.message_field(1)


class GetSteamNotificationsRequest(UnifiedMessage, um_name="SteamNotification.GetSteamNotifications"):
    include_hidden: bool = betterproto.bool_field(1)


class GetSteamNotificationsResponse(UnifiedMessage, um_name="SteamNotification.GetSteamNotifications"):
    notifications: "list[SteamNotificationData]" = betterproto.message_field(1)
    confirmation_count: int = betterproto.int32_field(2)


class HideNotificationNotification(UnifiedMessage, um_name="SteamNotification.HideNotification"):
    notification_ids: list[int] = betterproto.uint64_field(1)


class MarkNotificationsReadNotification(UnifiedMessage, um_name="SteamNotification.MarkNotificationsRead"):
    timestamp: int = betterproto.uint32_field(1)
    notification_type: int = betterproto.int32_field(2)
    notification_ids: list[int] = betterproto.uint64_field(3)
    mark_all_read: bool = betterproto.bool_field(4)


class SetPreferencesRequest(UnifiedMessage, um_name="SteamNotification.SetPreferences"):
    preferences: "list[SteamNotificationPreference]" = betterproto.message_field(1)


class SetPreferencesResponse(UnifiedMessage, um_name="SteamNotification.SetPreferences"):
    pass


@dataclass(eq=False, repr=False)
class SteamNotificationData(betterproto.Message):
    notification_id: int = betterproto.uint64_field(1)
    notification_targets: int = betterproto.uint32_field(2)
    notification_type: int = betterproto.int32_field(3)
    body_data: str = betterproto.string_field(4)
    read: bool = betterproto.bool_field(7)
    timestamp: int = betterproto.uint32_field(8)
    hidden: bool = betterproto.bool_field(9)


@dataclass(eq=False, repr=False)
class SteamNotificationPreference(betterproto.Message):
    notification_type: int = betterproto.int32_field(1)
    notification_targets: int = betterproto.uint32_field(2)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: econ_gcmessages.proto
# plugin: python-betterproto

from dataclasses import dataclass

import betterproto

from ....protobufs.msg import GCProtobufMessage
from ..enums import EMsg


@dataclass(eq=False, repr=False)
class GiftedItems(betterproto.Message):
    accountid: int = betterproto.uint32_field(1)
    giftdefindex: int = betterproto.uint32_field(2)
    max_gifts_possible: int = betterproto.uint32_field(3)
    num_eligible_recipients: int = betterproto.uint32_field(4)
    recipients_accountids: list[int] = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class ApplyAutograph(betterproto.Message):
    autograph_item_id: int = betterproto.uint64_field(1)
    item_item_id: int = betterproto.uint64_field(2)


class CasketItemAdd(GCProtobufMessage, msg=EMsg.CasketItemAdd):
    casket_item_id: int = betterproto.uint64_field(1)
    item_item_id: int = betterproto.uint64_field(2)


class CasketItemExtract(GCProtobufMessage, msg=EMsg.CasketItemExtract):
    casket_item_id: int = betterproto.uint64_field(1)
    item_item_id: int = betterproto.uint64_field(2)


class CasketItemLoadContents(GCProtobufMessage, msg=EMsg.CasketItemLoadContents):
    casket_item_id: int = betterproto.uint64_field(1)
    item_item_id: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class UserTrackTimePlayedConsecutively(betterproto.Message):
    state: int = betterproto.uint32_field(1)


class ItemCustomizationNotification(GCProtobufMessage, msg=EMsg.ItemCustomizationNotification):
    item_id: list[int] = betterproto.uint64_field(1)
    request: int = betterproto.uint32_field(2)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
from inspect import get_annotations
from typing import TYPE_CHECKING, Any, ClassVar, Final, cast, overload

from ..client import Client as Client_
from ..user import ClientUser as ClientUser_
from .state import GCState

if TYPE_CHECKING:
    from collections.abc import Mapping

    from typing_extensions import Self

    from ..app import App
    from ..enums import Language
    from ..trade import Inventory, Item
    from ..types.id import AppID
    from ..utils import cached_property

__all__ = ("Client",)


class ClientUser(ClientUser_):
    _state: GCState[Any]

    async def inventory(
        self, app: App, *, context_id: int | None = None, language: Language | None = None
    ) -> Inventory[Item[Self], Self]:
        return (  # type: ignore
            self._state.backpacks[app.id]
            if app.id in self._state.backpacks and self._state._gc_ready.is_set()
            else await super().inventory(app, context_id=context_id, language=language)
        )


class Client(Client_):
    _state: GCState[Any]
    _GC_HEART_BEAT: ClassVar = 30.0

    _ClientUserCls: ClassVar[type[ClientUser]] = ClientUser  # ideally one day this will be a generic
    _GC_BASES: Final[tuple[type[GCState[Any]], ...]] = ()
    _GC_APPS: Final[Mapping[AppID, App]] = {}
    user: cached_property[Self, ClientUser]  # type: ignore

    def __init_subclass__(cls) -> None:
        bases = [base for base in cls.__mro__ if issubclass(base, Client) and base is not Client]
        previous_state = None
        cls._GC_BASES = cast(  # type: ignore
            tuple[type[GCState[Any]], ...],
            tuple(
                dict.fromkeys(
                    previous_state := get_annotations(base, eval_str=True).get("_state", previous_state)
                    for base in reversed(bases)
                )
            ),
        )
        cls._GC_APPS = {  # type: ignore
            base._APP.id: base._APP for base in cls._GC_BASES
        }  # fmt: skip

        seen_methods = set[str]()
        for base in cls._GC_BASES:
            if (
                overridden_methods := seen_methods & base.__dict__.keys()
            ):  # How do we deal with subclasses of the same state overriding the method? Is anyone even going to do that?
                raise TypeError(f"Method(s) {', '.join(overridden_methods)} conflict inside of the state subclasses")
            seen_methods |= base.__dict__.keys()

    def _get_state(self, **options: Any) -> GCState[Any]:
        return type("GCState", self._GC_BASES, {})(self, **options)

    async def _ping_gc(self) -> None:
        await self._state.login_complete.wait()
        while not self.is_closed():
            for base in self._GC_BASES:
                if (msg := base._get_gc_message(self._state)) is not None:
                    await self._state.ws.send_gc_message(msg)
            await asyncio.sleep(self._GC_HEART_BEAT)

    def is_gc_ready(self) -> bool:
        return self._state._gc_ready.is_set()

    @overload
    async def login(
        self,
        username: str,
        password: str,
        *,
        shared_secret: str = ...,
        identity_secret: str = ...,
    ) -> None: ...

    @overload
    async def login(
        self,
        *,
        refresh_token: str,
        shared_secret: str = ...,
        identity_secret: str = ...,
    ) -> None: ...

    async def login(self, *args: Any, **kwargs: Any) -> None:
        await asyncio.gather(super().login(*args, **kwargs), self._ping_gc())

    async def _handle_ready(self) -> None:
        us = self._state._original_client_user_msg
        assert us is not None
        current_user = self._state.user
        user = self.http.user = self.__class__._ClientUserCls(self._state, us)
        user._friends = current_user._friends
        user._inventory_locks = current_user._inventory_locks
        self._state.user = user
        await super()._handle_ready()

    async def wait_until_gc_ready(self) -> None:
        """Wait for the :meth:`on_gc_ready` method to be dispatched.

        See Also
        --------
        :meth:`wait_for_ready`
        """
        await self._state._gc_ready.wait()

    # async def buy_item(self, def_index: int, price: int, def_indexes: list[int], prices: int) -> None:
    #     ...



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: econ_messages.proto
# plugin: python-betterproto

from dataclasses import dataclass

import betterproto

from ....protobufs.msg import GCProtobufMessage
from ..enums import EMsg


@dataclass(eq=False, repr=False)
class ApplyAutograph(betterproto.Message):
    autograph_item_id: int = betterproto.uint64_field(1)
    item_item_id: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class EconPlayerStrangeCountAdjustment(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    strange_count_adjustments: """list[EconPlayerStrangeCountAdjustmentCStrangeCountAdjustment]""" = (
        betterproto.message_field(2)
    )


@dataclass(eq=False, repr=False)
class EconPlayerStrangeCountAdjustmentCStrangeCountAdjustment(betterproto.Message):
    event_type: int = betterproto.uint32_field(1)
    item_id: int = betterproto.uint64_field(2)
    adjustment: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class RequestItemPurgatoryFinalizePurchase(betterproto.Message):
    item_ids: list[int] = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class RequestItemPurgatoryFinalizePurchaseResponse(betterproto.Message):
    result: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class RequestItemPurgatoryRefundPurchase(betterproto.Message):
    item_id: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class RequestItemPurgatoryRefundPurchaseResponse(betterproto.Message):
    result: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CraftingResponse(betterproto.Message):
    item_ids: list[int] = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class RequestStoreSalesData(betterproto.Message):
    version: int = betterproto.uint32_field(1)
    currency: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class RequestStoreSalesDataResponse(betterproto.Message):
    sale_price: "list[RequestStoreSalesDataResponsePrice]" = betterproto.message_field(1)
    version: int = betterproto.uint32_field(2)
    expiration_time: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class RequestStoreSalesDataResponsePrice(betterproto.Message):
    item_def: int = betterproto.uint32_field(1)
    price: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class RequestStoreSalesDataUpToDateResponse(betterproto.Message):
    version: int = betterproto.uint32_field(1)
    expiration_time: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class ToPingRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ToPingResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ToGetUserSessionServer(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class ToGetUserSessionServerResponse(betterproto.Message):
    server_steam_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class ToGetUserServerMembers(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    max_spectators: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class ToGetUserServerMembersResponse(betterproto.Message):
    member_account_id: list[int] = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class LookupMultipleAccountNames(betterproto.Message):
    accountids: list[int] = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class LookupMultipleAccountNamesResponse(betterproto.Message):
    accounts: "list[LookupMultipleAccountNamesResponseAccount]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class LookupMultipleAccountNamesResponseAccount(betterproto.Message):
    accountid: int = betterproto.uint32_field(1)
    persona: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ToGrantSelfMadeItemToAccount(betterproto.Message):
    item_def_index: int = betterproto.uint32_field(1)
    accountid: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class ToThankedByNewUser(betterproto.Message):
    new_user_accountid: int = betterproto.uint32_field(1)
    thanked_user_accountid: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class ShuffleCrateContents(betterproto.Message):
    crate_item_id: int = betterproto.uint64_field(1)
    user_code_string: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class QuestObjectiveProgress(betterproto.Message):
    quest_id: int = betterproto.uint64_field(1)
    quest_attrib_index: int = betterproto.uint32_field(2)
    delta: int = betterproto.uint32_field(3)
    owner_steamid: int = betterproto.fixed64_field(4)


@dataclass(eq=False, repr=False)
class QuestObjectivePointsChange(betterproto.Message):
    quest_id: int = betterproto.uint64_field(1)
    owner_steamid: int = betterproto.fixed64_field(4)
    update_base_points: bool = betterproto.bool_field(5)
    points_0: int = betterproto.uint32_field(6)
    points_1: int = betterproto.uint32_field(7)
    points_2: int = betterproto.uint32_field(8)


@dataclass(eq=False, repr=False)
class QuestCompleteRequest(betterproto.Message):
    quest_id: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class QuestCompleted(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class QuestObjectiveRequestLoanerItems(betterproto.Message):
    quest_id: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class QuestObjectiveRequestLoanerResponse(betterproto.Message):
    pass


class CraftCollectionUpgrade(GCProtobufMessage, msg=EMsg.CraftCollectionUpgrade):
    item_id: list[int] = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class CraftHalloweenOffering(betterproto.Message):
    tool_id: int = betterproto.uint64_field(1)
    item_id: list[int] = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CraftCommonStatClock(betterproto.Message):
    tool_id: int = betterproto.uint64_field(1)
    item_id: list[int] = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class QuestDiscardRequest(betterproto.Message):
    quest_id: int = betterproto.uint64_field(1)
